import psycopg2
import polars as pl
from autodla.engine.data_conversion import DataTransformer, DataConversion
from autodla.engine.interfaces import QueryBuilder_Interface
from autodla.engine.object import primary_key
from datetime import date, datetime
from typing import List, Optional
from autodla.utils.logger import logger
from uuid import UUID
import os
import asyncio
from autodla.utils.watchdog import Watchdog
from autodla.dbs.memorydb import MemoryDB
from autodla.utils.df_tools import df_comparator, ensure_dtype_equality
import traceback

DATETIME_FORMAT = "%Y-%m-%d %H:%M:%S"
if "DATETIME_FORMAT" in os.environ:
    DATETIME_FORMAT = os.environ.get("DATETIME_FORMAT")
POSTGRES_USER = 'postgres'
if "AUTODLA_POSTGRES_USER" in os.environ:
    POSTGRES_USER = os.environ.get("AUTODLA_POSTGRES_USER")
POSTGRES_PASSWORD = 'password'
if "AUTODLA_POSTGRES_PASSWORD" in os.environ:
    POSTGRES_PASSWORD = os.environ.get("AUTODLA_POSTGRES_PASSWORD")
POSTGRES_URL = 'localhost'
if "AUTODLA_POSTGRES_HOST" in os.environ:
    POSTGRES_URL = os.environ.get("AUTODLA_POSTGRES_HOST")
POSTGRES_DB = 'my_db'
if "AUTODLA_POSTGRES_DB" in os.environ:
    POSTGRES_DB = os.environ.get("AUTODLA_POSTGRES_DB")
DB_FLUSH_TIME = 60
if "AUTODLA_DB_FLUSH_TIME" in os.environ:
    DB_FLUSH_TIME = int(os.environ.get("AUTODLA_DB_FLUSH_TIME"))

CONNECTION_URL = "postgresql://{}:{}@{}/{}".format(
    POSTGRES_USER,
    POSTGRES_PASSWORD,
    POSTGRES_URL,
    POSTGRES_DB
)


def to_name(st: str) -> str:
    if "." not in st:
        st = "public." + st
    return st


class PostgresQueryBuilder(QueryBuilder_Interface):
    def __init__(self, data_transformer: DataTransformer) -> None:
        self._data_transformer = data_transformer

    def select(
        self,
        from_table: str,
        columns: List[str],
        where: Optional[str] = None,
        limit: Optional[int] = 10,
        order_by: Optional[str] = None,
        group_by: Optional[list[str]] = None,
        offset: Optional[int] = None
    ) -> str:
        qry = "SELECT " + ", ".join(columns) + " FROM " + to_name(from_table)
        if where:
            qry += " WHERE " + where
        if order_by:
            qry += " ORDER BY " + order_by
        if limit:
            qry += " LIMIT " + str(limit)
        if offset:
            qry += " OFFSET " + str(offset)
        return qry

    def insert(
        self,
        into_table: str,
        values: List[dict]
    ) -> str:
        qry = "".join([
            "INSERT INTO ",
            to_name(into_table),
            " (" + ", ".join(values[0].keys()),
            ") VALUES "
        ])
        qry += ", ".join([
            f"({', '.join(
                [self._data_transformer.convert_data(v) for v in d.values()]
            )})" for d in values
        ])
        return qry

    def update(
        self,
        table: str,
        values: dict,
        where: str
    ) -> str:
        qry = (
            f"UPDATE {to_name(table)} SET"
            f"{', '.join([
                (
                    f'{k.upper()} = {self._data_transformer.convert_data(v)}'
                ) for k, v in values.items()
            ])}"
            f" WHERE {where}"
        )
        return qry

    def delete(
        self,
        table: str,
        where: str
    ) -> str:
        qry = f"DELETE FROM {to_name(table)} WHERE {where}"
        return qry

    def create_table(
        self,
        table_name: str,
        schema: dict,
        if_exists: bool = False
    ) -> str:
        if_exists_st = "IF EXISTS" if if_exists else ""
        items = [f'{k} {v}' for k, v in schema.items()]
        qry = "CREATE TABLE {} {} ({});".format(
            if_exists_st,
            to_name(table_name),
            ', '.join(items)
        )
        return qry

    def drop_table(
        self,
        table_name: str,
        if_exists: bool = False
    ) -> str:
        if_exists_st = "IF EXISTS" if if_exists else ""
        qry = f"DROP TABLE {if_exists_st} {to_name(table_name)};"
        return qry


class PostgresDataTransformer(DataTransformer):
    TYPE_DICT = {
        UUID: DataConversion("UUID", lambda x: f"'{x}'"),
        primary_key: DataConversion("UUID", lambda x: f"'{x}'"),
        type(None): DataConversion('', lambda x: "NULL"),
        int: DataConversion('INTEGER'),
        float: DataConversion("REAL"),
        str: DataConversion("TEXT", lambda x: f"'{x}'"),
        bool: DataConversion(
            "BOOL", lambda x: {True: "TRUE", False: "FALSE"}[x]
        ),
        date: DataConversion(
            "DATE", lambda x: f"'{x.year}-{x.month}-{x.day}'"
        ),
        datetime: DataConversion(
            "TIMESTAMP", lambda x: f"'{x.strftime(DATETIME_FORMAT)}'"
        ),
    }
    OPERATOR_DICT = {
        "numeric": {
            'Eq': "=",
            'NotEq': "<>",
            'Lt': "<",
            'LtE': "<=",
            'Gt': ">",
            'GtE': ">=",
            'In': "IN",
            'NotIn': "NOT IN",
            'Is': "IS",
            'IsNot': "IS NOT"
        },
        "binary": {
            "Add": lambda x, y: f'{x} + {y}',
            "Sub": lambda x, y: f'{x} - {y}',
            "Mult": lambda x, y: f'{x} * {y}',
            "Div": lambda x, y: f'{x} / {y}',
            "FloorDiv": lambda x, y: f'FLOOR({x} / {y})',
            "Mod": lambda x, y: f'{x} % {y}',
            "Pow": lambda x, y: f'POWER({x},{y})'
        },
        "boolean": {
            "And": 'AND',
            "Or": 'OR',
        },
        "unary": {
            "Not": 'NOT'
        }
    }
    NODE_COMPATIBILITY = {
        primary_key: UUID,
        UUID: primary_key
    }


class PostgresDB(MemoryDB):

    def __init__(self, connection_url=CONNECTION_URL) -> None:
        self.__querys_executed_postgres = 0
        super().__init__()
        self.__connection_url = connection_url
        self.__pg_connection: Optional[psycopg2.extensions.connection] = None
        self.__pg_dt = PostgresDataTransformer()
        self.__pg_query = PostgresQueryBuilder(self.__pg_dt)
        self.__last_sync = datetime.now()
        self.watchdog = Watchdog(DB_FLUSH_TIME, self.sync)
        self.__atached = False
        self.__mid_sync = False
        self.connect()

    def connect(self) -> bool:
        """
        Connects to the PostgreSQL database.
        """
        try:
            self.__pg_connection = (
                psycopg2.connect(self.__connection_url)
            )
            self._execute(
                "update pg_cast set castcontext='a' where "
                "casttarget = 'boolean'::regtype;"
            )
            return True
        except psycopg2.OperationalError as e:
            logger.error(f"Failed to connect to PostgreSQL: {e}")
        return False

    @property
    def usage_metrics(self):
        """
        Returns the usage metrics of the MemoryDB connection.
        This is a placeholder method and should be implemented in subclasses.
        """
        return {
            "postgres": self.__querys_executed_postgres,
            **super().usage_metrics
        }

    async def poll_watchdog(self):
        while True:
            await asyncio.sleep(1)
            self.watchdog.check()

    def execute(self, statement):
        if self.__atached:
            self.watchdog.reset()
            if not self.__mid_sync:
                datetime_diff = (
                    datetime.now() - self.__last_sync
                ).total_seconds()
                if datetime_diff > DB_FLUSH_TIME:
                    self.sync()
        res = super().execute(statement)
        return res

    def attach(self, objects):
        if self.__pg_connection is None:
            if not self.connect():
                logger.error(
                    "Failed to connect to PostgreSQL, cannot execute query."
                )
                return None
        logger.debug("Attaching objects to PostgreSQL...\n")
        super().attach(objects)
        for table in self._tables:
            self.ensure_table(
                table,
                self._table_schemas[table],
                current_data_schema=self._get_table_definition(table))
        self.__atached = True
        self.sync()
        logger.debug("Attached objects to PostgreSQL...\n")

    def sync(self) -> None:
        if self.__pg_connection is None:
            if not self.connect():
                logger.error(
                    "Failed to connect to PostgreSQL, cannot execute query."
                )
                return
        if self.__mid_sync:
            logger.debug("Sync already in progress, skipping...")
            return
        logger.debug("Syncing to PostgreSQL...")
        self.__mid_sync = True
        self.__last_sync = datetime.now()

        external_snapshot: dict[str, pl.DataFrame] = (
            self.snapshot_tables_external()
        )
        internal_snapshot: dict[str, pl.DataFrame] = (
            self.snapshot_tables()
        )

        final_snapshot = {}
        for table_name in internal_snapshot:
            in_df = internal_snapshot[table_name]
            ex_df = external_snapshot[table_name]

            if in_df.is_empty() and ex_df.is_empty():
                continue
            if in_df.is_empty():
                final_snapshot[table_name] = ex_df
                continue
            if ex_df.is_empty():
                final_snapshot[table_name] = in_df
                continue

            in_df, ex_df = ensure_dtype_equality(in_df, ex_df)

            final_df = pl.concat([in_df, ex_df], how="vertical").sort(
                "dla_modified_at", descending=True
            )
            final_df = final_df.group_by(
                "dla_object_id", maintain_order=True).first()

            final_snapshot[table_name] = final_df

        if final_snapshot is None or len(final_snapshot) == 0:
            logger.debug("No changes to sync.")
            self.__mid_sync = False
            return

        internal_querys = []
        external_querys = []
        for table_name, df in final_snapshot.items():

            def get_change_querys(
                    snapshot: dict[str, pl.DataFrame],
                    query_executor: QueryBuilder_Interface
            ) -> List[str]:
                out = []
                add_df = df_comparator(df, snapshot, method="insert")
                update_df = df_comparator(df, snapshot, method="update")
                logger.debug(
                    f"add_df: {query_executor.__class__} {add_df}")
                logger.debug(
                    f"update_df: {query_executor.__class__} {update_df}")
                if not add_df.is_empty():
                    out.append(
                        query_executor.insert(
                            into_table=table_name, values=add_df.to_dicts()
                        )
                    )
                if not update_df.is_empty():
                    for row in update_df.to_dicts():
                        where_clause = (
                            f"dla_object_id = "f"'{row['dla_object_id']}'"
                        )
                        out.append(
                            query_executor.update(
                                table=table_name,
                                values=row,
                                where=where_clause)
                        )
                return out
            internal_querys += get_change_querys(
                internal_snapshot[table_name], self.query)
            external_querys += get_change_querys(
                external_snapshot[table_name], self.__pg_query)
        if len(internal_querys) > 0:
            self.execute(internal_querys)
        if len(external_querys) > 0:
            self._execute(external_querys)
        logger.debug("Syncing to PostgreSQL completed.")
        self.__mid_sync = False

    def exit(self) -> None:
        self.sync()
        return super().exit()

    def snapshot_tables_external(self) -> dict[str, pl.DataFrame]:
        out: dict[str, pl.DataFrame] = {}
        for table, schema in self._table_schemas.items():
            qry = self.__pg_query.select(
                from_table=f"{table} t",
                columns=[f"t.{i}" for i in schema],
                limit=None)
            res = self._execute([qry], commit=False)
            if res is None:
                res = pl.DataFrame(
                    data=None,
                    schema=[k.lower() for k in schema]
                )
            out[table] = res
        return out

    def _get_table_definition(self, table_name) -> dict[str, type]:
        if "." in table_name:
            table_name = table_name.split(".")[-1]
        res = self._execute([self.__pg_query.select(
            from_table='INFORMATION_SCHEMA.COLUMNS',
            columns=["column_name", "data_type"],
            limit=None,
            where=f"table_name = '{table_name}'"
        )], commit=False)
        out = {}
        if res is not None:
            res = res.to_dicts()
            conversion_dict = {
                "boolean": "bool",
                "timestamp without time zone": "timestamp"
            }
            for row in res:
                if row['data_type'] in conversion_dict:
                    row['data_type'] = conversion_dict[row['data_type']]
                out[row['column_name'].upper()] = (
                    self.__pg_dt.get_type_from_sql_type(row["data_type"])
                )
        return out

    def _execute(
        self,
        statements: list,
        commit=True
    ) -> Optional[pl.DataFrame | None]:
        self.__querys_executed_postgres += 1
        statements = (
            statements
            if isinstance(statements, list)
            else [statements]
        )
        with self.__pg_connection.cursor() as cursor:
            try:
                out = None
                for statement in statements:
                    statement = self.normalize_statement(statement)
                    logger.debug(
                        '{"running": "PostgresDB._execute", '
                        '"statement": "' + statement + '"}'
                    )
                    try:
                        cursor.execute(statement)
                    except Exception as e:
                        raise ValueError(
                            f"Error executing query:\n< {statement} "
                            f">\nError: <{traceback.format_exc()} {e}>"
                        )
                    try:
                        rows = cursor.fetchall()
                    except psycopg2.ProgrammingError:
                        continue
                    if rows:
                        schema = [desc[0] for desc in cursor.description]
                        out = pl.DataFrame(rows, schema=schema, orient='row')
                        out.columns = [col.lower() for col in out.columns]
                        logger.debug(
                            '{"running": "PostgresDB._execute", "result": "'
                            f'{out}'
                            '"}'
                        )
                return out
            except Exception as e:
                logger.error(f"{traceback.format_exc()} {e}")
                return None
            finally:
                if commit:
                    self.__pg_connection.commit()



from fastapi import APIRouter, Depends, FastAPI, HTTPException, Request, status
from fastapi.responses import FileResponse
from typing import (
    Annotated,
    Callable,
    Optional,
    Type,
)
from fastapi.security import OAuth2PasswordRequestForm
import json
from autodla.engine.interfaces import DB_Connection_Interface
from autodla.engine.object import Object
from autodla.engine.web_connection import EndpointMaker, WebConnection
from autodla.engine.lambda_conversion import json_to_lambda_str
import inspect
import asyncio
from contextlib import asynccontextmanager


class FastApiEndpointMaker(EndpointMaker):
    @classmethod
    def list(cls, object: Type[Object]) -> Callable:
        async def read_object(
            limit: int = 10,
            skip: int = 0,
            filter: Optional[str] = None
        ):
            if filter is None:
                res = object.all(limit, skip=skip)
            else:
                filter_dict = json.loads(filter)
                lambda_st = json_to_lambda_str(filter_dict)
                res = object.filter(lambda_st, limit, skip=skip)
            out = []
            for i in res:
                out.append(i.to_dict())
            return out
        return read_object

    @classmethod
    def get(cls, object: Type[Object]) -> Callable:
        async def get_object_id(
            id_param: str
        ):
            res = object.get_by_id(id_param)
            if res is None:
                return HTTPException(400, f'{object.__name__} not found')
            return res.to_dict()
        return get_object_id

    @classmethod
    def get_history(cls, object: Type[Object]) -> Callable:
        async def get_object_history_id(
            id_param: str
        ):
            res = object.get_by_id(id_param)
            if res is None:
                return HTTPException(400, f'{object.__name__} not found')
            return res.history()
        return get_object_history_id

    @classmethod
    def table(cls, object: Type[Object]) -> Callable:
        async def read_table(
            limit: int = 10,
            skip: int = 0,
            only_current: bool = True,
            only_active: bool = True
        ):
            if not (res := object.get_table_res(
                limit=limit,
                skip=skip,
                only_current=only_current,
                only_active=only_active
            )):
                return HTTPException(400, f'{object.__name__} not found')
            return res.to_dicts()
        return read_table

    @classmethod
    def new(cls, object_class: Type[Object]) -> Callable:
        async def create_object(obj: Object):
            n = object_class.new(**obj.model_dump())
            return n.to_dict()
        return create_object

    @classmethod
    def edit(cls, object: Type[Object]) -> Callable:
        async def edit_object(id_param, data: dict):
            if (obj := object.get_by_id(id_param)) is None:
                raise HTTPException(
                    status_code=404,
                    detail=f"{object.__name__} with id {id_param} not found"
                )
            obj.update(**data)
            return obj.to_dict()
        return edit_object

    @classmethod
    def delete(cls, object: Type[Object]) -> Callable:

        async def delete_object(id_param: str):
            if (obj := object.get_by_id(id_param)) is None:
                raise HTTPException(
                    status_code=404,
                    detail=f"{object.__name__} with id {id_param} not found"
                )
            obj.delete()
            return {"status": "done"}
        return delete_object


class FastApiWebConnection(WebConnection):
    def __init__(
        self,
        app: FastAPI,
        db: DB_Connection_Interface,
        setup_autodla_web: bool = True,
        admin_endpoints_prefix: str = '/autodla-admin'
    ):
        self.app = app
        self.db = db
        if poll_watchdog := getattr(db, "poll_watchdog", None):
            @asynccontextmanager
            async def lifespan(app):
                asyncio.create_task(poll_watchdog())
                yield
            app.lifespan = lifespan  # type: ignore

        self.admin_endpoints_prefix = admin_endpoints_prefix
        super().__init__(FastApiEndpointMaker(), setup_autodla_web)

    def admin_endpoint_validate(self, func):
        if not inspect.iscoroutinefunction(func):
            raise Exception("Admin endpoint must be an async function")
        return func

    def normalize_endpoint(self, func):
        sig = inspect.signature(func)
        orig_params = list(sig.parameters.values())
        if orig_params and orig_params[0].name == "request":
            return func

        async def new_func(request: Request, *args, **kwargs):
            return await func(request, *args, **kwargs)
        new_func.__signature__ = sig.replace(
            parameters=[
                inspect.Parameter(
                    "request",
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    annotation=Request
                )
            ] + list(sig.parameters.values())
        )
        new_func.__name__ = func.__name__
        new_func.__doc__ = func.__doc__
        sig = inspect.signature(new_func)
        orig_params = list(sig.parameters.values())
        return new_func

    async def extract_token(self, *args, **kwargs):
        request = kwargs.get("request") or (args[0] if args else None)
        auth_header = request.headers.get("Authorization")
        if not auth_header or not auth_header.startswith("Bearer "):
            raise HTTPException(status_code=401, detail="Invalid token format")
        token = auth_header.split(" ")[1]
        return token

    def setup_admin_endpoints(self):
        self.admin_router = APIRouter(prefix="/admin", tags=["autodla_admin"])

        @self.admin_router.post("/token")
        async def login(
            form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
        ):
            current_token = self.login(form_data.username, form_data.password)
            return {"access_token": current_token, "token_type": "bearer"}

        @self.admin_router.get('/get_json_schema')
        @self.admin_endpoint
        async def get_schema():
            return self.db.get_json_schema()

        self.app.include_router(
            self.admin_router,
            prefix=self.admin_endpoints_prefix
        )

    def create_crud_router(
        self,
        object_class: Type[Object],
        prefix=None,
        tags=[],
        auth_wrapper=None
    ) -> APIRouter:
        if prefix is None:
            prefix = f"/{object_class.__name__}"
        if tags == []:
            tags = [f"autodla_{object_class.__name__}"]
        router = APIRouter(prefix=prefix, tags=tags)
        endpoints = [
            ("list", "get", "/list"),
            ("get", "get", "/get/{id_param}"),
            ("get_history", "get", "/get_history/{id_param}"),
            ("table", "get", "/table"),
            ("new", "post", "/new"),
            ("edit", "put", "/edit/{id_param}"),
            ("delete", "delete", "/delete/{id_param}")
        ]
        for func_name, method, path in endpoints:
            endpoint_func = getattr(
                self.endpoint_maker, func_name
            )(object_class)
            router_wrapper = getattr(router, method)
            if auth_wrapper is not None:
                endpoint_func = auth_wrapper(endpoint_func)
            router_args = {
                "path": path
            }
            if method == "new":
                router_args["response_model"] = object_class.__name__
            router_wrapper(path)(endpoint_func)
        return router

    def create_static_router(self):
        static_temp_dir = self.static_temp_dir
        web_router = APIRouter(prefix="/autodla-web", tags=["autodla_web"])
        sub_directories = ['', 'assets/']
        for sub_directory in sub_directories:

            @web_router.get('/' + sub_directory + '{filename}')
            async def static_files(filename: str = 'index.html'):
                return FileResponse(
                    f'{static_temp_dir}/{sub_directory}{filename}'
                )

        @web_router.get('/')
        async def static_home():
            return FileResponse(f'{static_temp_dir}/index.html')
        return web_router

    def setup_autodla_web_endpoints(self):
        web_router = self.create_static_router()
        self.app.include_router(web_router)
        for cls_type in self.db.classes:
            print(f"Creating CRUD router for {cls_type.__name__}")
            r = self.create_crud_router(
                cls_type,
                auth_wrapper=self.admin_endpoint
            )
            self.app.include_router(r, prefix=self.admin_endpoints_prefix)

    @classmethod
    def unauthorized_handler(cls):
        raise HTTPException(
            status_code=status.HTTP_401_UNAUTHORIZED,
            detail="Not authenticated",
            headers={"WWW-Authenticate": "Bearer"},
        )

    @classmethod
    def invalid_admin_credentials_handler(cls):
        raise HTTPException(
            status_code=400,
            detail="Incorrect username or password"
        )



from datetime import datetime, timedelta
import random
from typing import List


def select_random(x: list):
    """
    Selects a random element from the provided list.
    :param x: List from which to select a random element.
    :return: A randomly selected element from the list.
    """
    return x[int(random.random() * len(x))]


class PossibleValues:
    name: List[str] = [
        'James', 'John', 'Robert', 'Michael', 'William', 'David', 'Richard',
        'Joseph', 'Thomas', 'Charles', 'Mary', 'Patricia', 'Jennifer', 'Linda',
        'Elizabeth', 'Barbara', 'Susan', 'Jessica', 'Sarah', 'Karen', 'Emma',
        'Olivia', 'Noah', 'Liam', 'Sophia', 'Ava', 'Isabella', 'Mia',
        'Abigail', 'Emily', 'Alexander', 'Ethan', 'Daniel', 'Matthew', 'Aiden',
        'Henry', 'Joseph', 'Jackson', 'Samuel', 'Sebastian', 'Sofia',
        'Charlotte', 'Amelia', 'Harper', 'Evelyn', 'Aria', 'Scarlett', 'Grace',
        'Chloe', 'Victoria'
    ]
    age: List[int] = [i for i in range(10, 30)]
    mass: List[float] = [(0.5 + (0.1*i)) for i in range(10, 60)]
    created_at: List[datetime] = [
        datetime.now() - timedelta(
            days=i, minutes=10+20*random.random()
        ) for i in range(10, 60)
    ]


class DataGenerator:
    @staticmethod
    def name() -> str:
        return select_random(PossibleValues.name)

    @staticmethod
    def age() -> int:
        return select_random(PossibleValues.age)

    @staticmethod
    def mass() -> float:
        return select_random(PossibleValues.mass)

    @staticmethod
    def created_at() -> datetime:
        return select_random(PossibleValues.created_at)



import logging
import sys
import os

logger = logging.getLogger("autodla")
logger.setLevel(logging.DEBUG)
logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

VERBOSE = os.environ.get("AUTODLA_SQL_VERBOSE", "false").lower() in (
    "true", "1", "yes"
)
logger.disabled = not VERBOSE



import time
from typing import Callable


class Watchdog:
    def __init__(self, timeout: float, on_timeout: Callable):
        self.timeout = timeout
        self.on_timeout = on_timeout
        self._last_reset = time.time()
        self._timeout_triggered = False
        self._enabled = True

    def reset(self):
        self._last_reset = time.time()
        self._timeout_triggered = False

    def stop(self):
        self._enabled = False

    def check(self):
        if not self._enabled:
            return
        if time.time() - self._last_reset > self.timeout:
            if not self._timeout_triggered:
                self._timeout_triggered = True
                self.on_timeout()



from __future__ import annotations
from json import JSONEncoder
from dataclasses import field, _MISSING_TYPE
from datetime import datetime
from types import NoneType
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Type,
    TypeVar,
    Union,
    get_origin,
    ClassVar,
    Literal,
    get_args
)
import uuid
import polars as pl
from autodla.engine.lambda_conversion import lambda_to_sql
from pydantic import GetCoreSchemaHandler
from pydantic_core import CoreSchema, core_schema, PydanticUndefinedType
from autodla.utils.logger import logger
from autodla.engine.interfaces import (
    DB_Connection_Interface,
    Object_Interface,
    Table_Interface,
    primary_key_Interface
)
import warnings
warnings.filterwarnings('error')


def _default(self, obj):
    return getattr(obj.__class__, "to_json", _default.default)(obj)


_default.default = JSONEncoder().default  # type: ignore
JSONEncoder.default = _default  # type: ignore

primary_key_type = TypeVar('primary_key_type', bound='primary_key')


class primary_key(primary_key_Interface):
    @classmethod
    def generate(cls: Type[primary_key_type]) -> primary_key_type:
        return cls(str(uuid.uuid4()))

    def is_valid(self):
        try:
            uuid.UUID(self)
            return True
        except ValueError:
            return False

    @staticmethod
    def auto_increment():
        return field(default_factory=lambda: primary_key.generate())

    def __eq__(self, value):
        if isinstance(value, str):
            return super().__eq__(value)
        if isinstance(value, uuid.UUID):
            return uuid.UUID(self) == value

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        return core_schema.no_info_after_validator_function(cls, handler(str))

    def __hash__(self):
        return super().__hash__()


def dla_dict(
        operation: Literal["INSERT", "UPDATE", "DELETE"],
        modified_at: datetime = datetime.now(),
        modified_by: str = "SYSTEM",
        is_current: bool = False,
        is_active: bool = True
):
    def out():
        return {
            'DLA_object_id': primary_key.generate(),
            'DLA_modified_at': modified_at,
            'DLA_operation': operation,
            'DLA_modified_by': modified_by,
            'DLA_is_current': is_current,
            'DLA_is_active': is_active
        }
    return out


class Table(Table_Interface):
    def __init__(
            self,
            table_name: str,
            schema: dict,
            db: Optional[DB_Connection_Interface | None] = None
    ):
        self.schema = schema
        if db:
            table_name_res = db.get_table_name(table_name)
            self.table_name = table_name
            self.__table_alias = table_name_res.alias
            self.set_db(db)

    @property
    def db(self) -> DB_Connection_Interface:
        db = self.__db
        if db is None:
            raise ValueError("DB not defined")
        return db

    def set_db(self, db: DB_Connection_Interface) -> None:
        if db is None:
            raise ValueError("DB not defined")
        self.__db = db
        self.__db.ensure_table(self.table_name, self.schema, save=True)

    def get_all(
            self,
            limit: Optional[int] = 10,
            only_current: bool = True,
            only_active: bool = True,
            skip: int = 0
    ) -> Optional[pl.DataFrame]:
        conditions = ["TRUE"]
        if only_current:
            conditions.append("DLA_is_current = true")
        if only_active:
            conditions.append("DLA_is_active = true")
        where_st = " AND ".join(conditions)
        qry = self.db.query.select(
            from_table=''
            + f'{self.__db.get_table_name(self.table_name).name} '
            + f'{self.__table_alias}',
            columns=[f'{self.__table_alias}.{i}' for i in list(
                self.schema.keys())],
            where=where_st,
            limit=limit,
            offset=skip
        )
        return self.db.execute(qry)

    def filter(
            self,
            l_func: Callable,
            limit: Optional[int] = 10,
            only_current: bool = True,
            only_active: bool = True,
            skip: int = 0
    ) -> Optional[pl.DataFrame]:
        conditions = [
            lambda_to_sql(
                self.schema,
                l_func,
                self.__db.data_transformer,
                alias=self.__table_alias
            )
        ]
        if only_current:
            conditions.append("DLA_is_current = true")
        if only_active:
            conditions.append("DLA_is_active = true")
        where_st = " AND ".join(conditions)
        qry = self.db.query.select(
            from_table=f'{
                self.__db.get_table_name(self.table_name).name
            } {
                self.__table_alias
            }',
            columns=[f'{self.__table_alias}.{i}' for i in list(
                self.schema.keys())],
            where=where_st,
            limit=limit,
            offset=skip
        )
        return self.db.execute(qry)

    def insert(self, data: dict) -> None:
        qry = self.db.query.insert(
            self.__db.get_table_name(self.table_name).name, [data])
        self.db.execute(qry)

    def update(self, l_func: Callable, data: dict) -> None:
        where_st = lambda_to_sql(
            self.schema, l_func, self.__db.data_transformer, alias="")
        update_data = {f'{key}': value for key, value in data.items()}
        qry = self.db.query.update(
            f'{self.__db.get_table_name(self.table_name).name}',
            where=where_st,
            values=update_data
        )
        self.db.execute(qry)

    def delete_all(self) -> None:
        qry = self.db.query.delete(
            self.__db.get_table_name(self.table_name).name, "TRUE")
        self.db.execute(qry)


class Object(Object_Interface):
    __table: ClassVar[Optional[Table | None]] = None
    __dependencies: ClassVar[dict[str, Object_Interface.ObjectDependency]] = (
        field(
            default_factory=dict
        )
    )
    identifier_field: ClassVar[str] = "id"
    __objects_list: ClassVar[List['Object']] = list()
    __objects_map: ClassVar[dict[str, 'Object']] = dict()

    @classmethod
    def delete_all(cls) -> None:
        cls.__objects_list = []
        cls.__objects_map = {}
        if cls.__table is not None:
            cls.__table.delete_all()

    @classmethod
    def set_db(cls, db: DB_Connection_Interface) -> None:
        schema = cls.get_types()
        dependencies: dict[str, Object.ObjectDependency] = {}
        common_fields = {
            'DLA_object_id': {
                "type": uuid.UUID
            },
            'DLA_modified_at': {
                "type": datetime
            },
            'DLA_operation': {
                "type": str
            },
            'DLA_modified_by': {
                "type": str
            },
            'DLA_is_current': {
                "type": bool
            },
            'DLA_is_active': {
                'type': bool
            }
        }
        for k, i in schema.items():
            if 'depends' in i:
                table_name = '' \
                    + f"{cls.__name__.lower()}"\
                    + f"__{k}__"\
                    + f"{i['depends'].__name__.lower()}"
                dependencies[k] = Object.ObjectDependency(
                    is_list=i.get("is_list") is True,
                    is_value=False,
                    type=i['depends'],
                    table=Table(
                        table_name,
                        {
                            "connection_id": {
                                "type": primary_key
                            },
                            "first_id": {
                                "type": primary_key
                            },
                            "second_id": {
                                "type": primary_key
                            },
                            "list_index": {
                                "type": int
                            }, **common_fields
                        },
                        db
                    )
                )
            elif 'is_list' in i and i["type"] is not None:
                table_name = f"{cls.__name__.lower()}__{k}"
                dependencies[k] = Object.ObjectDependency(
                    is_list=i.get("is_list") is True,
                    is_value=True,
                    type=i["type"],
                    table=Table(
                        table_name,
                        {
                            "connection_id": {
                                "type": primary_key
                            },
                            "first_id": {
                                "type": primary_key
                            },
                            "value": {
                                "type": i["type"]
                            },
                            "list_index": {
                                "type": int
                            }, **common_fields
                        },
                        db
                    )
                )
        for key in dependencies.keys():
            del schema[key]
        cls.__table = Table(cls.__name__.lower(), {
                            **schema, **common_fields}, db)
        cls.__dependencies = dependencies

    @classmethod
    def get_types(cls) -> dict[str, Object_Interface.TypeDictionary]:
        out = {}
        fields = cls.model_fields
        for i in fields:
            if (get_origin(fields[i].annotation) == ClassVar):
                continue
            type_out: Object_Interface.TypeDictionary = {}
            tp: Optional[Type | None] = fields[i].annotation
            ori, arg = get_origin(tp), get_args(tp)
            if ori == Union:
                if arg[1] == NoneType:
                    type_out["nullable"] = True
                    tp = arg[0]
                    ori, arg = get_origin(tp), get_args(tp)
            if (
                    type_out.get('nullable') is True
                    and fields[i].default is not None):
                raise TypeError(
                    'Field with type Optional must initialize to None')
            if (
                    type_out.get('nullable') is not True
                    and fields[i].default is None):
                raise TypeError(
                    'Field initialized to None must be of type Optional')
            if type(fields[i].default) not in [
                    _MISSING_TYPE,
                    PydanticUndefinedType]:
                type_out["default"] = fields[i].default
            if type(fields[i].default_factory) not in [
                    _MISSING_TYPE,
                    PydanticUndefinedType]:
                type_out["default_factory"] = fields[i].default_factory
            if ori == list:
                tp = arg[0]
                ori, arg = get_origin(tp), get_args(tp)
                type_out["is_list"] = True
            if tp is not None:
                if issubclass(tp, Object):
                    type_out["depends"] = tp
            type_out["type"] = tp
            out[i] = type_out
        return out

    @classmethod
    def _update_individual(
        cls,
        data_inp: dict[str, Any]
    ) -> Optional['Object_Interface']:
        logger.debug(f"UPDATE INDIVIDUAL: {cls} {data_inp}")
        data: dict[str, Any] = {}
        for k, v in data_inp.items():
            if not k.upper().startswith("DLA_"):
                data[k] = v
        found = cls.__objects_map.get(data[cls.identifier_field])
        try:
            cls.model_validate(data)
        except Exception as e:
            logger.error(
                f"Validation error for {cls.__name__} with data {data}: {e}")
            return None
        if found is not None:
            found.__dict__.update(data)
            return found
        obj = cls(**data)
        cls.__objects_list.append(obj)
        cls.__objects_map[obj[cls.identifier_field]] = obj
        return obj

    @classmethod
    def _update_info(
        cls,
        filter: Optional[Callable[[Any], bool]] = None,
        limit: Optional[int] = 10,
        skip: int = 0,
        only_current: bool = True,
        only_active: bool = True
    ) -> list['Object_Interface']:
        if cls.__table is None:
            return []
        if filter is None:
            res = cls.__table.get_all(
                limit, only_current, only_active, skip=skip)
        else:
            res = cls.__table.filter(
                filter, limit, only_current, only_active, skip=skip)
        obj_lis: list[dict[str, Any]] = []
        if res is not None:
            obj_lis = res.to_dicts()
        if obj_lis == []:
            return []
        if res is None:
            return []
        id_list = res[cls.identifier_field].to_list()

        table_results: dict[str, Optional[pl.DataFrame]] = {}
        dep_tables_required_ids: dict[
            str, Object_Interface.DependencyRequiredIds] = {}
        for k, v in cls.__dependencies.items():
            if v.is_value:
                continue
            dep_table_: Table_Interface = v.table
            table_results[k] = dep_table_.filter(
                lambda x: x.first_id in id_list,
                None, only_current=only_current,
                only_active=only_active
            )
            ids: set[str] = set()
            df: Optional[pl.DataFrame] = table_results[k]
            if df is not None:
                ids = set(df['second_id'].to_list())
            t_name: str = v.type.__class__.__name__
            if t_name not in dep_tables_required_ids:
                tp: 'Type[Object_Interface]' = v.type
                dep_tables_required_ids[t_name] = (
                    Object_Interface.DependencyRequiredIds(
                        type=tp, ids=ids
                    )
                )
            else:
                row = dep_tables_required_ids[t_name]
                row.ids = row.ids.union(ids)

        dep_tables: dict[str, dict[str, "Object_Interface"]] = {}
        for k_, v_ in dep_tables_required_ids.items():
            l: list[str] = list(v_.ids)
            tp_: 'Type[Object_Interface]' = v_.type
            id_field = tp_.identifier_field
            dep_tables[k_] = {}
            if len(l) == 0:
                continue
            filter_res = tp_.filter(lambda x: x[id_field] in l)
            for obj in filter_res:
                dep_tables[k_][getattr(obj, tp_.identifier_field)] = obj

        out = []
        for obj_dic in obj_lis:
            dep_key: str
            for dep_key in cls.__dependencies.keys():
                dep: Object.ObjectDependency = cls.__dependencies[dep_key]
                if dep.is_value:
                    obj_id: primary_key = obj_dic[cls.identifier_field]
                    dep_table: Table_Interface = dep.table
                    table_df: Optional[pl.DataFrame] = dep_table.filter(
                        lambda x: x.first_id == obj_id
                    )
                    table_res: list[Any] = []
                    if table_df is not None:
                        table_res = table_df['value'].to_list()
                    obj_dic[dep_key] = table_res
                    continue
                df = table_results[dep_key]
                val_lis = []
                t_name = dep.type.__class__.__name__
                if df is not None and len(df) > 0:
                    lis = df.filter(
                        df['first_id'] == obj[cls.identifier_field]
                    )[
                        'second_id'].to_list()
                    for row in lis:
                        val = dep_tables[t_name].get(str(row))
                        if val is not None:
                            val_lis.append(val)
                obj_dic[dep_key] = val_lis
                if not cls.__dependencies[dep_key].is_list:
                    if obj[dep_key] != []:
                        obj_dic[dep_key] = obj[dep_key][0]
                    else:
                        obj_dic[dep_key] = None
            update_obj: Optional['Object_Interface'] = cls._update_individual(
                obj_dic)
            if update_obj is not None:
                out.append(update_obj)
        return out

    @classmethod
    def new(cls, **kwargs) -> 'Object':
        if cls.__table is None:
            raise ImportError('DB not defined')
        if cls.identifier_field in kwargs:
            del kwargs[cls.identifier_field]
        out = cls(**kwargs)
        data = out.to_dict()
        for i in cls.__dependencies:
            del data[i]
        dla_data = dla_dict("INSERT", is_current=True)
        cls.__table.insert({**data, **dla_data()})
        for field_i, v in cls.__dependencies.items():
            if v.is_list:
                new_rows = []
                if v.is_value:
                    for idx, i in enumerate(getattr(out, field_i)):
                        new_rows.append({
                            'connection_id': primary_key.generate(),
                            "first_id": out[cls.identifier_field],
                            "value": i,
                            "list_index": idx,
                            **dla_data()
                        })
                else:
                    for idx, i in enumerate(getattr(out, field_i)):
                        new_rows.append({
                            'connection_id': primary_key.generate(),
                            "first_id": out[cls.identifier_field],
                            "second_id": getattr(i, v.type.identifier_field),
                            "list_index": idx,
                            **dla_data()
                        })
                for j in new_rows:
                    v.table.insert(j)
            else:
                val = getattr(out, field_i)
                if val is not None:
                    v.table.insert({
                        'connection_id': primary_key.generate(),
                        "first_id": out[cls.identifier_field],
                        "second_id": val[v.type.identifier_field],
                        "list_index": 0,
                        **dla_data()
                    })
        cls.__objects_map[str(out[cls.identifier_field])] = out
        cls.__objects_list.append(out)
        return out

    def history(self) -> dict[str, list[dict[str, Any]]]:
        if self.__table is None:
            return {}
        self_res = self.__table.filter(
            lambda x: x[self.identifier_field] == getattr(
                self, self.identifier_field
            ),
            limit=None,
            only_active=False,
            only_current=False
        )
        out: dict[str, list[dict[str, Any]]] = {
            "self": self_res.to_dicts() if self_res is not None else [],
        }
        for k, v in self.__dependencies.items():
            dep_res = v.table.filter(lambda x: x.first_id == getattr(
                self, self.identifier_field), limit=None, only_active=False,
                only_current=False)
            out[k] = dep_res.to_dicts() if dep_res is not None else []
        return out

    def update(self, **kwargs) -> None:
        if self.__table is None:
            return
        data = {}
        for key in self.to_dict():
            if key in kwargs:
                data[key] = kwargs[key]
            elif key not in self.__class__.__dependencies:
                data[key] = getattr(self, key)
        dla_data_insert = dla_dict("UPDATE", is_current=True)
        for key, value in kwargs.items():
            if key in self.__dependencies:
                del data[key]
                dependency = self.__dependencies[key]
                dependency.table.update(
                    lambda x: x.first_id == self[self.identifier_field], {
                        'DLA_is_current': False})
                new_rows = []
                if dependency.is_list:
                    if dependency.is_value:
                        for idx, i in enumerate(value):
                            new_rows.append({
                                'connection_id': primary_key.generate(),
                                "first_id": self[self.identifier_field],
                                "value": i,
                                "list_index": idx,
                                **dla_data_insert()
                            })
                    else:
                        for idx, i in enumerate(value):
                            new_rows.append({
                                'connection_id': primary_key.generate(),
                                "first_id": self[self.identifier_field],
                                "second_id": i[
                                    dependency.type.identifier_field],
                                "list_index": idx,
                                **dla_data_insert()
                            })
                else:
                    if value is not None:
                        new_rows.append({
                            'connection_id': primary_key.generate(),
                            "first_id": self[self.identifier_field],
                            "second_id": value[
                                dependency.type.identifier_field],
                            "list_index": 0,
                            **dla_data_insert()
                        })
                for j in new_rows:
                    dependency.table.insert(j)
            setattr(self, key, value)
        self.__table.update(lambda x: x[self.identifier_field] == self[
            self.identifier_field], {
            'DLA_is_current': False})
        self.__table.insert({**data, **dla_data_insert()})

    def delete(self) -> None:
        if self.__table is None:
            return
        data = {}
        for key in self.__class__.model_fields:
            data[key] = getattr(self, key)
        dla_data_delete = dla_dict("DELETE", is_current=True, is_active=False)
        for key, dependency in self.__dependencies.items():
            del data[key]
            dependency.table.update(lambda x: x.first_id == self[
                self.identifier_field], {'DLA_is_current': False})
            value = getattr(self, key)
            if value is None:
                continue
            new_rows = []
            if dependency.is_list:
                if dependency.is_value:
                    for idx, i in enumerate(value):
                        new_rows.append({
                            'connection_id': primary_key.generate(),
                            "first_id": self[self.identifier_field],
                            "value": i,
                            "list_index": idx,
                            **dla_data_delete()
                        })
                else:
                    for idx, i in enumerate(value):
                        new_rows.append({
                            'connection_id': primary_key.generate(),
                            "first_id": self[self.identifier_field],
                            "second_id": i[dependency.type.identifier_field],
                            "list_index": idx,
                            **dla_data_delete()
                        })
            else:
                if value is not None:
                    new_rows.append({
                        'connection_id': primary_key.generate(),
                        "first_id": self[self.identifier_field],
                        "second_id": value[dependency.type.identifier_field],
                        "list_index": 0,
                        **dla_data_delete()
                    })
            for j in new_rows:
                dependency.table.insert(j)
        self.__table.update(lambda x: x[self.identifier_field] == self[
            self.identifier_field], {
            'DLA_is_current': False})
        self.__table.insert({**data, **dla_data_delete()})

    @classmethod
    def all(cls, limit=10, skip=0) -> list[Object_Interface]:
        out = cls._update_info(limit=limit, skip=skip)
        return out

    @classmethod
    def filter(cls, lambda_f, limit=10, skip=0) -> list[Object_Interface]:
        out = cls._update_info(filter=lambda_f, limit=limit, skip=skip)
        return out

    @classmethod
    def get_by_id(cls, id_param) -> Optional["Object"]:
        cls._update_info(
            lambda x: x[cls.identifier_field] == id_param, limit=1, skip=0)
        return cls.__objects_map.get(id_param)

    @classmethod
    def get_table_res(
        cls, limit=10, skip=0, only_current=True,
        only_active=True
    ) -> Optional[pl.DataFrame]:
        if cls.__table is None:
            return None
        return cls.__table.get_all(
            limit=limit, only_current=only_current,
            only_active=only_active, skip=skip)

    def to_dict(self) -> dict[str, Any]:
        return self.model_dump()

    def to_json(self) -> str:
        return self.model_dump_json()

    def __getitem__(self, item: str) -> Any:
        return getattr(self, item)



import builtins
import ast
import inspect
from dataclasses import dataclass
from typing import Any, Callable, Optional, Type
from autodla.engine.interfaces import (
    DataTransformer_Interface,
    MethodArgument
)
from datetime import datetime
from autodla.utils.logger import logger


class LambdaFinder(ast.NodeVisitor):
    def __init__(self):
        self.found = None

    def find(self, tree):
        self.visit(tree)
        return self.found

    def generic_visit(self, node):
        if isinstance(node, ast.Lambda):
            self.found = node
            return
        ast.NodeVisitor.generic_visit(self, node)


class CallableType(object):
    def __init__(self, func: Callable):
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)

    def __repr__(self):
        return f"Callable({self.func.__name__})"


def lambda_to_ast(lambda_func):
    try:
        if type(lambda_func) is str:
            source = lambda_func
        else:
            source = inspect.getsource(lambda_func).strip()
    except (IOError, TypeError) as e:
        raise TypeError(f"Failed to get lambda source: {str(e)}")
    root_tree = ast.parse(source)
    return LambdaFinder().find(root_tree)


def lambda_to_text(lambda_func):
    lambda_node = lambda_to_ast(lambda_func)
    return ast.unparse(lambda_node)


@dataclass
class NodeReturn:
    st: str
    tp: Optional[Type]
    eval: Optional[Any] = None


def pn(node):
    logger.debug("")
    if isinstance(node, ast.AST):
        logger.debug(ast.dump(node))
    else:
        logger.debug(node)
    logger.debug("")


class LambdaToSql(ast.NodeVisitor):
    def __init__(
        self,
        root: ast.Lambda,
        schema: dict[str, Type],
        data_transformer: DataTransformer_Interface,
        ctx_vars: dict[str, Any] = {},
        alias: str = 'x'
    ):
        self.ctx_vars = ctx_vars
        self.schema = schema
        self.root = root
        self.data_transformer = data_transformer
        self.alias = alias

    def evaluate_node(
        self,
        node: ast.AST
    ) -> Any:
        return eval(str(ast.unparse(node)), self.ctx_vars)

    def repr(
        self,
        obj: Any
    ) -> str:
        st = repr(obj)
        if callable(obj):
            st = obj.__name__
            if hasattr(obj, '__self__'):
                st = self.repr(obj.__self) + '.' + st
        return st

    def obj_to_node(
        self,
        obj: Any
    ) -> ast.AST:
        st = self.repr(obj)
        if type(obj) in [int, float, bool, str, datetime]:
            return ast.Constant(value=obj)
        found_stmt = ast.parse(st).body[0]
        if isinstance(found_stmt, ast.Expr):
            return found_stmt.value
        raise ValueError("Error turning value to node")

    def evaluate_and_parse_node(
        self,
        node: ast.AST
    ) -> NodeReturn:
        out = self.obj_to_node(self.evaluate_node(node))
        return self.parse_node(out)

    def node_compatibility(
        self,
        node1: NodeReturn,
        node2: NodeReturn
    ) -> bool:
        if isinstance(node1, type(node2)) or isinstance(node2, type(node1)):
            return True
        if node1 is None or node2 is None:
            return True
        if isinstance(node1, list):
            if all([self.node_compatibility(arg, node2) for arg in node1]):
                return True
        elif isinstance(node2, list):
            if all([self.node_compatibility(arg, node1) for arg in node2]):
                return True
        return False

    def parse_node(
        self,
        node: ast.AST
    ) -> NodeReturn:
        match node:
            case ast.Subscript():
                caller = self.parse_node(node.value)
                slice_node = self.parse_node(node.slice)
                attr = slice_node.st if (
                    slice_node.eval) is None else slice_node.eval
                if caller.eval is None:
                    if attr not in self.schema:
                        raise AttributeError(
                            "invalid attribute for {}: '{}'".format(
                                caller.st,
                                attr
                            ))
                    if caller.st == "":
                        return NodeReturn(attr, self.schema.get(attr))
                    return NodeReturn(
                        f'{caller.st}.{attr}',
                        self.schema.get(attr)
                    )
                if slice_node.eval is None:
                    raise ValueError(f'invalid slice node: {slice_node.st}')
                val = getattr(caller.eval, slice_node.eval)
                if val is None:
                    raise AttributeError(f"attribute not found: '{attr}'")
                if callable(val):
                    return NodeReturn(attr, CallableType, val)
                return self.evaluate_and_parse_node(node)
            case ast.IfExp():
                condition = self.parse_node(node.test)
                condition_value = self.parse_node(node.body)
                else_value = self.parse_node(node.orelse)
                if all([
                    getattr(arg, 'eval', None) is not None for arg in [
                        condition, condition_value, else_value
                    ]
                ]):
                    return self.evaluate_and_parse_node(node)
                if condition.tp != bool:
                    raise TypeError('condition is not a valid bool')
                if not self.node_compatibility(condition_value, else_value):
                    raise TypeError('values for if should be of same type')
                st = '(CASE WHEN {} THEN {} ELSE {} END)'.format(
                    condition.st,
                    condition_value.st,
                    else_value.st
                )
                return NodeReturn(st, condition_value.tp)
            case ast.Call():
                args = [self.parse_node(arg) for arg in node.args]
                func_name = ""
                func = None
                match node.func:
                    case ast.Name():
                        func_name = node.func.id
                        f = self.ctx_vars.get(func_name)
                        if f is not None:
                            if callable(f):
                                func = f
                            if type(f) is type(int) and all(
                                [arg.eval is not None for arg in args]
                            ):
                                res = f(*[arg.eval for arg in args])
                                return self.parse_node(self.obj_to_node(res))
                    case ast.Attribute():
                        caller = self.parse_node(node.func.value)
                        func_name = node.func.attr
                        if caller.eval is not None:
                            func = getattr(caller.eval, func_name)
                if func is not None and all([
                    arg.eval is not None for arg in args
                ]):
                    return self.evaluate_and_parse_node(node)

                if caller is None:
                    found = self.data_transformer.get_method(None, func_name)
                    if found is not None:
                        st, tp = found(
                            *[MethodArgument(arg.st, arg.tp) for arg in args])
                        return NodeReturn(st, tp)
                elif caller.tp is not None:
                    found = self.data_transformer.get_method(
                        caller.tp, func_name)
                    if found is not None:
                        method_list: list[MethodArgument] = []
                        for arg in args:
                            if arg.tp is None:
                                raise TypeError(
                                    f"argument {arg.st} is not defined"
                                )
                            else:
                                method_list.append(MethodArgument(
                                    arg.st,
                                    arg.tp
                                ))
                        st, tp = found(
                            MethodArgument(caller.st, caller.tp),
                            *method_list
                        )
                        return NodeReturn(st, tp)
                raise ValueError(f'function {func_name} not found')

            case ast.List():
                values = [self.parse_node(i) for i in node.elts]
                st_out = f'({", ".join([node_i.st for node_i in values])})'
                eval_list = [node_i.eval for node_i in values]
                if all([node_i is not None for node_i in eval_list]):
                    return NodeReturn(st_out, list, eval_list)
                return NodeReturn(st_out, list)
            case ast.UnaryOp():
                op = self.data_transformer.get_operator("unary", node.op)
                child = self.parse_node(node.operand)
                if child.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({op} {child.st})',
                    bool
                )
            case ast.BoolOp():
                values = [self.parse_node(i) for i in node.values]
                for i in values:
                    if i.tp != bool:
                        raise TypeError(
                            f"expression '{i.st}' is not a valid Boolean")
                op = self.data_transformer.get_operator("boolean", node.op)
                out = f' {op} '.join([i.st for i in values])
                if all([arg.eval is not None for arg in values]):
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({out})',
                    bool
                )
            case ast.BinOp():
                left = self.parse_node(node.left)
                right = self.parse_node(node.right)
                bin_func = self.data_transformer.get_operator(
                    "binary",
                    node.op
                )
                if isinstance(bin_func, str):
                    raise ValueError(
                        f"binary operator '{bin_func}' not found"
                    )
                if left.tp != right.tp:
                    raise TypeError(
                        f"invalid operation between {left.tp} and {right.tp}")
                if left.eval is not None and right.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({bin_func(left.st, right.st)})',
                    left.tp
                )
            case ast.Name():
                if node.id == 'x':
                    return NodeReturn(self.alias, str)
                value = self.ctx_vars.get(node.id)
                if value is not None:
                    try:
                        transformed_value = self.data_transformer.convert_data(
                            value)
                    except Exception as e:
                        logger.error(
                            f"Error converting data for {node.id}: {e}"
                        )
                        transformed_value = node.id
                    return NodeReturn(transformed_value, type(value), value)
                raise ValueError(f"'{node.id}' is not defined")
            case ast.Constant():
                return NodeReturn(
                    self.data_transformer.convert_data(node.value),
                    type(node.value),
                    node.value
                )
            case ast.Compare():
                left = self.parse_node(node.left)
                op = self.data_transformer.get_operator("numeric", node.ops[0])
                right = self.parse_node(node.comparators[0])
                if not self.node_compatibility(left, right):
                    raise TypeError(
                        "invalid comparission between {} and {}".format(
                            left.tp,
                            right.tp
                        )
                    )
                if left.eval is not None and right.eval is not None:
                    return self.evaluate_and_parse_node(node)
                return NodeReturn(
                    f'({left.st} {op} {right.st})', bool
                )
            case ast.Attribute():
                caller = self.parse_node(node.value)
                if caller.eval is None:
                    if node.attr not in self.schema:
                        raise AttributeError(
                            f"invalid attribute for x: '{node.attr}'")
                    if caller.st == "":
                        return NodeReturn(
                            node.attr,
                            self.schema.get(node.attr)
                        )
                    return NodeReturn(
                        f'{caller.st}.{node.attr}',
                        self.schema.get(node.attr)
                    )
                val = getattr(caller.eval, node.attr)
                if val is None:
                    raise AttributeError(f"attribute not found: '{node.attr}'")
                if callable(val):
                    return NodeReturn(node.attr, CallableType, val)
                return self.evaluate_and_parse_node(node)
            case _:
                try:
                    evaluation = self.evaluate_node(node)
                    new_node = ast.Constant(value=evaluation)
                    return self.parse_node(new_node)
                except Exception as e:
                    logger.error(f"Error evaluating node: {e}")
                    logger.debug(ast.dump(node))
                    raise ValueError(
                        f'Invalid node type: {type(node).__name__}')

    def transform(self) -> NodeReturn:
        definition_error = False
        if len(self.root.args.args) != 1:
            definition_error = True
        if list(self.root.args.args)[0].arg != 'x':
            definition_error = True
        if definition_error:
            raise SyntaxError("lambda definition should be -> 'lambda x:'")
        out = self.parse_node(self.root.body)
        return out


def get_context_from_lamba(lambda_func) -> dict[str, Any]:
    file = lambda_func.__code__.co_filename
    line = lambda_func.__code__.co_firstlineno
    found = None
    for frame in inspect.stack():
        if frame.filename == file and frame.lineno == line:
            found = frame
            break
    if found is None:
        raise ValueError('frame not found')
    return {**vars(builtins), **found.frame.f_locals}


def lambda_to_sql(
        schema: dict[str, Type],
        lambda_func: Callable[[Any], bool],
        data_transformer: DataTransformer_Interface,
        ctx_vars={},
        alias='x'
) -> str:
    if not type(lambda_func) is str:
        ctx_vars = get_context_from_lamba(lambda_func)
    lambda_node = lambda_to_ast(lambda_func)
    out = LambdaToSql(lambda_node, schema, data_transformer=data_transformer,
                      ctx_vars=ctx_vars, alias=alias).transform()
    return out.st


def json_to_lambda_str(json_condition: dict) -> str:
    """
    Transforms a SQL-inspired JSON condition to a
    Python lambda string representation.

    Args:
        json_condition (dict): A condition object that can be:
            - Simple: {"field": "age", "operator": "gt", "value": 10}
            - Complex: {"and": [condition1, condition2, ...]}
                or {"or": [condition1, condition2, ...]}

    Returns:
        str: A string representation of the lambda function
    """
    # Check if this is a complex condition with AND/OR
    if "and" in json_condition:
        sub_conditions = [json_to_lambda_str(
            cond) for cond in json_condition["and"]]
        return f"lambda x: {
            ' and '.join(f'({cond})' for cond in sub_conditions)
        }"

    elif "or" in json_condition:
        sub_conditions = [json_to_lambda_str(
            cond) for cond in json_condition["or"]]
        return f"lambda x: {
            ' or '.join(f'({cond})' for cond in sub_conditions)
        }"

    # Handle negation
    elif "not" in json_condition:
        sub_condition = json_to_lambda_str(json_condition["not"])
        # Extract the condition part (after "lambda x: ")
        cond_part = sub_condition.split("lambda x: ", 1)[1]
        return f"lambda x: not ({cond_part})"

    # Handle simple condition
    elif all(k in json_condition for k in ["field", "operator"]):
        field = json_condition.get("field")
        operator = json_condition.get("operator")
        value = json_condition.get("value")

        # Map of operators to Python comparison operators
        operator_map = {
            "eq": "==",
            "neq": "!=",
            "gt": ">",
            "gte": ">=",
            "lt": "<",
            "lte": "<=",
            "in": "in",
            "nin": "not in"
        }

        if operator not in operator_map:
            raise ValueError(f"Unsupported operator: {operator}")

        # Generate the lambda function string
        op_str = operator_map[operator]

        # Format the value appropriately
        if isinstance(value, str):
            formatted_value = f"'{value}'"
        elif isinstance(value, list):
            # Format each element in the list
            formatted_elements = []
            for elem in value:
                if isinstance(elem, str):
                    formatted_elements.append(f"'{elem}'")
                else:
                    formatted_elements.append(str(elem))
            formatted_value = f"[{', '.join(formatted_elements)}]"
        else:
            formatted_value = str(value)

        return f"lambda x: x.{field} {op_str} {formatted_value}"
    else:
        raise ValueError(f"Invalid condition format: {json_condition}")



import sqlite3
from typing import List, Optional
import polars as pl
from autodla.engine.data_conversion import DataTransformer
from autodla.engine.db import DB_Connection, TableName
from autodla.engine.interfaces import (
    QueryBuilder_Interface,
    DataConversion
)
from autodla.engine.object import primary_key
from autodla.utils.logger import logger
from datetime import date, datetime
from uuid import UUID, uuid4
import os
import time
import traceback

import threading

DATETIME_FORMAT: Optional[str] = "%Y-%m-%d %H:%M:%S"
if "DATETIME_FORMAT" in os.environ:
    DATETIME_FORMAT = os.environ.get("DATETIME_FORMAT")


def to_name(st):
    if "." in st:
        return st
    if st[0] != '"' and st[-1] != '"':
        st = f'"{st}"'
    return st


class MemoryQueryBuilder(QueryBuilder_Interface):

    def __init__(self, data_transformer: DataTransformer) -> None:
        self._data_transformer = data_transformer

    def select(
        self,
        from_table: str,
        columns: List[str],
        where: Optional[str] = None,
        limit: Optional[int] = 10,
        order_by: Optional[str] = None,
        group_by: Optional[list[str]] = None,
        offset: Optional[int] = None
    ) -> str:
        qry = "SELECT " + ", ".join(
            [to_name(i) for i in columns]
        ) + " FROM " + to_name(from_table)
        if where:
            qry += " WHERE " + where
        if order_by:
            qry += " ORDER BY " + order_by
        if limit:
            qry += " LIMIT " + str(limit)
        if offset:
            qry += " OFFSET " + str(offset)
        return qry

    def insert(
        self,
        into_table: str,
        values: List[dict]
    ) -> str:
        qry = "INSERT INTO " + to_name(
            into_table
        ) + " (" + ", ".join(values[0].keys()) + ") VALUES "
        qry += ", ".join([
            f"({', '.join([
                self._data_transformer.convert_data(v) for v in d.values()
            ])})" for d in values
        ])
        return qry

    def update(
        self,
        table: str,
        values: dict,
        where: str
    ) -> str:
        qry = f"UPDATE {to_name(table)} SET {', '.join([f'{k.upper()} = {
            self._data_transformer.convert_data(v)
        }' for k, v in values.items()])} WHERE {where}"
        return qry

    def delete(
        self,
        table: str,
        where: str
    ) -> str:
        qry = f"DELETE FROM {to_name(table)} WHERE {where}"
        return qry

    def create_table(
        self,
        table_name: str,
        schema: dict,
        if_exists: bool = False
    ) -> str:
        if_exists_st = "IF NOT EXISTS" if if_exists else ""
        items = [f'{k} {v}' for k, v in schema.items()]
        qry = f"CREATE TABLE {if_exists_st} {to_name(table_name)} ({
            ', '.join(items)});"
        return qry

    def drop_table(
        self,
        table_name: str,
        if_exists: bool = False
    ) -> str:
        if_exists_st = "IF EXISTS" if if_exists else ""
        qry = f"DROP TABLE {if_exists_st} {to_name(table_name)};"
        return qry


class MemoryDataTransformer(DataTransformer):
    TYPE_DICT = {
        UUID: DataConversion("TEXT", lambda x: f"'{x}'"),
        primary_key: DataConversion("TEXT", lambda x: f"'{x}'"),
        type(None): DataConversion('', lambda x: "NULL"),
        int: DataConversion('INTEGER'),
        float: DataConversion('REAL'),
        str: DataConversion('TEXT', lambda x: f"'{x}'"),
        bool: DataConversion('INTEGER', lambda x: "1" if x else "0"),
        date: DataConversion(
            'TEXT',
            lambda x: f"'{x.year}-{x.month}-{x.day}'"
        ),
        datetime: DataConversion(
            'TEXT',
            lambda x: f"'{x.strftime(DATETIME_FORMAT)}'"
        )
    }
    OPERATOR_DICT = {
        "numeric": {
            'Eq': "=",
            'NotEq': "!=",
            'Lt': "<",
            'LtE': "<=",
            'Gt': ">",
            'GtE': ">=",
            'In': "IN",
            'NotIn': "NOT IN",
            'Is': "IS",
            'IsNot': "IS NOT"
        },
        "binary": {
            "Add": lambda x, y: f'{x} + {y}',
            "Sub": lambda x, y: f'{x} - {y}',
            "Mult": lambda x, y: f'{x} * {y}',
            "Div": lambda x, y: f'{x} / {y}',
            "FloorDiv": lambda x, y: f'({x} / {y})',
            "Mod": lambda x, y: f'{x} % {y}',
            "Pow": lambda x, y: f'POWER({x},{y})'
        },
        "boolean": {
            "And": 'AND',
            "Or": 'OR'
        },
        "unary": {
            "Not": 'NOT'
        }
    }
    NODE_COMPATIBILITY = {
        primary_key: UUID,
        UUID: primary_key
    }


class MemoryDB(DB_Connection):
    def __init__(self):
        self.query_queue = []
        self.result_queue = {}
        self.lock = threading.Lock()
        self.db_status = {"connected": False}
        threading.Thread(target=self.pooling_queue).start()
        while True:
            time.sleep(0.1)
            with self.lock:
                if self.db_status["connected"]:
                    break
        dt = MemoryDataTransformer()
        self.tables = {}
        super().__init__(dt, MemoryQueryBuilder(dt))
        self.__querys_executed_memory = 0

    @property
    def usage_metrics(self):
        """
        Returns the usage metrics of the MemoryDB connection.
        This is a placeholder method and should be implemented in subclasses.
        """
        return {"memorydb": self.__querys_executed_memory}

    def exit(self) -> None:
        with self.lock:
            if self.query_queue is not None:
                logger.debug("Closing MemoryDB connection...")
                self.query_queue = None

    def pooling_queue(self) -> None:
        self.__db_connection = sqlite3.connect(":memory:")
        with self.lock:
            self.db_status["connected"] = True
        while True:
            if self.query_queue is None:
                break
            if self.query_queue:
                with self.lock:
                    query = self.query_queue.pop(0)
                    res = self._execute_memory(
                        **{
                            "statement_input": query["statement"],
                        }
                    )
                    self.result_queue[query["execution_id"]] = res
            else:
                time.sleep(0.1)

    def get_table_name(self, table_name: str) -> TableName:
        return TableName(
            name=f'"{table_name.upper()}"',
            alias=f'"{table_name.lower()}"'
        )

    def attach(self, objects):
        logger.debug("Attaching objects to MemoryDB...\n")
        super().attach(objects)
        logger.debug("Attached objects to MemoryDB...\n")

    def get_table_definition(self, table_name) -> dict[str, type]:
        try:
            res_exec = self.execute(
                f"PRAGMA table_info('{table_name.split('.')[-1]}')"
            )
            res = res_exec.to_dicts() if res_exec is not None else []
            if not res:
                {}
            out = {}
            for row in res:
                tp = self.data_transformer.get_type_from_sql_type(row["type"])
                out[row['name'].upper()] = tp
            return out
        except Exception as e:
            logger.error(
                f"Error getting table definition for {table_name}: {e}"
            )
            return {}

    def execute(
        self,
        query: str
    ) -> Optional[pl.DataFrame]:
        self.__querys_executed_memory += 1
        execution_id = str(uuid4())
        with self.lock:
            if self.query_queue is None:
                return super().execute(query)
            self.query_queue.append({
                "statement": query,
                "execution_id": execution_id
            })
        while True:
            with self.lock:
                if execution_id in self.result_queue:
                    result = self.result_queue.pop(execution_id)
                    return result
            time.sleep(0.1)

    def _execute_memory(self, statement_input: str) -> Optional[pl.DataFrame]:
        logger.debug(
            '{"running": "MemoryDB.execute", "statement": "' + str(
                statement_input) + '"}'
        )
        statements = statement_input if isinstance(
            statement_input, list) else [statement_input]
        cursor = self.__db_connection.cursor()
        try:
            out = None
            for statement in statements:
                statement = self.normalize_statement(statement)
                try:
                    cursor.execute(statement)
                except Exception as e:
                    raise ValueError(
                        "Error executing query:\n<{}>\nError: <{} {}>".format(
                            statement, traceback.format_exc(), e
                        )
                    )
                rows = cursor.fetchall()
                if rows:
                    schema = [desc[0] for desc in cursor.description]
                    out = pl.DataFrame(rows, schema=schema, orient='row')
                    out.columns = [col.lower() for col in out.columns]
                    logger.debug(
                        '{"running": "MemoryDB.execute", "result": "'
                        + str(out) + '"}')
            return out
        except Exception as e:
            logger.error(f"{e}")
            return None
        finally:
            self.__db_connection.commit()

    def snapshot_tables(self) -> dict[str, pl.DataFrame]:
        out = {}
        for table, schema in self._table_schemas.items():
            qry = self.query.select(
                from_table=self.get_table_name(table).name,
                columns=list(schema.keys()),
                limit=None
            )
            res = self.execute(qry)
            if res is None:
                res = pl.DataFrame(
                    data=None,
                    schema=[k.lower() for k in schema]
                )
            out[table] = res
        return out



from typing import Literal, Tuple
import polars as pl


def ensure_dtype_equality(
    df1: pl.DataFrame,
    df2: pl.DataFrame
) -> Tuple[pl.DataFrame, pl.DataFrame]:
    if set(df1.columns) != set(df2.columns):
        raise ValueError("DataFrames have different columns.")
    for col in df1.columns:
        expected_dtype: pl.DataType = df1[col].dtype
        if df1[col].dtype == df2[col].dtype:
            expected_dtype = df1[col].dtype
        if df1[col].dtype == pl.Null and df2[col].dtype != pl.Null:
            expected_dtype = df2[col].dtype
        if df2[col].dtype == pl.Null and df1[col].dtype != pl.Null:
            expected_dtype = df1[col].dtype
        if (df1[col].dtype == pl.String) or (df2[col].dtype == pl.String):
            expected_dtype = pl.String()
        if expected_dtype == pl.Null:
            continue

        # Ensure both DataFrames have the same dtype for the column
        if df1[col].dtype != expected_dtype:
            df1 = df1.with_columns(pl.col(col).cast(expected_dtype))
        if df2[col].dtype != expected_dtype:
            df2 = df2.with_columns(pl.col(col).cast(expected_dtype))
    return df1, df2


def df_comparator(
    df1: pl.DataFrame,
    df2: pl.DataFrame,
    method: Literal["insert", "update"]
) -> pl.DataFrame:
    df1, df2 = ensure_dtype_equality(df1, df2)
    out = pl.DataFrame(data=None, schema=df1.columns)
    match method:
        case "insert":
            if df1.is_empty() or df2.is_empty():
                return df1
            return df1.join(df2, on="dla_object_id", how="anti")
        case "update":
            if df1.is_empty() or df2.is_empty():
                return out
            joined = df1.join(
                df2,
                on="dla_object_id",
                how="inner",
                suffix="_ext"
            )

            # Build expression to detect differences
            # between columns (excluding ID)
            diff_exprs = []
            for col in df1.columns:
                if col != "dla_object_id":
                    diff_exprs.append(pl.col(col) != pl.col(f"{col}_ext"))

            if diff_exprs:
                # Combine all diff expressions using logical OR
                condition = diff_exprs[0]
                for expr in diff_exprs[1:]:
                    condition = condition | expr

                # Filter rows where any column differs
                modified_rows = joined.filter(condition)

                if not modified_rows.is_empty():
                    # Optionally, keep only the columns
                    # from the current snapshot
                    return modified_rows.select(df1.columns)
    return out



from datetime import datetime
import os
from functools import wraps
from typing import Callable, Optional
import hashlib
import uuid
from abc import ABC, abstractmethod


def hash(st):
    return hashlib.sha256(st.encode()).hexdigest()


AUTODLAWEB_USER: Optional[str] = 'autodla'
if "AUTODLAWEB_USER" in os.environ:
    AUTODLAWEB_USER = os.environ.get("AUTODLAWEB_USER")
AUTODLAWEB_PASSWORD = hash('password')
if "AUTODLAWEB_PASSWORD" in os.environ:
    AUTODLAWEB_PASSWORD = hash(os.environ.get("AUTODLAWEB_PASSWORD"))


def generate_token():
    return hash(str(uuid.uuid4()))


class EndpointMaker(ABC):
    @classmethod
    @abstractmethod
    def list(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def get(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def get_history(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def table(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def new(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def edit(cls, object) -> Callable:
        ...

    @classmethod
    @abstractmethod
    def delete(cls, object) -> Callable:
        ...


class WebConnection(ABC):
    def __init__(
        self,
        endpoint_maker: EndpointMaker,
        setup_autodla_web: bool = True
    ) -> None:
        self.endpoint_maker = endpoint_maker
        self.current_token = generate_token()
        self.setup_admin_endpoints()
        if setup_autodla_web:
            self.setup_autodla_web_static_files()
            self.setup_autodla_web_endpoints()

    @abstractmethod
    def setup_admin_endpoints(self) -> None:
        ...

    def setup_autodla_web_static_files(self) -> None:
        import os
        from importlib import resources as impresources
        from .. import static as staticdir
        import tempfile
        import shutil
        temp_dir = tempfile.mkdtemp()
        static_package_dir = impresources.files(staticdir)
        static_temp_dir = os.path.join(temp_dir, 'static')
        os.makedirs(static_temp_dir, exist_ok=True)

        def copy_dir_recursively(source_dir, dest_dir):
            os.makedirs(dest_dir, exist_ok=True)
            for item in source_dir.iterdir():
                dest_path = os.path.join(dest_dir, item.name)
                if item.is_file():
                    with item.open('rb') as src, open(dest_path, 'wb') as dst:
                        shutil.copyfileobj(src, dst)
                elif item.is_dir():
                    copy_dir_recursively(item, dest_path)
        copy_dir_recursively(static_package_dir, static_temp_dir)
        self.static_temp_dir = static_temp_dir

    @abstractmethod
    def setup_autodla_web_endpoints(self) -> None:
        ...

    # -- Admin token management

    _current_admin_tokens: dict[str, datetime] = {}

    def create_new_admin_token(self) -> str:
        new_token = generate_token()
        self._current_admin_tokens[new_token] = datetime.now()
        return new_token

    def validate_token(self, token: str) -> bool:
        now = datetime.now()
        for token, timestamp in self._current_admin_tokens.items():
            if (now - timestamp).total_seconds() > 60 * 60 * 3:
                del self._current_admin_tokens[token]
        if self._current_admin_tokens.get(token, None) is not None:
            self.unauthorized_handler()
        return True

    def login(self, username: str, password: str):
        if (username != AUTODLAWEB_USER
                or hash(password) != AUTODLAWEB_PASSWORD):
            self.invalid_admin_credentials_handler()
        return self.create_new_admin_token()

    def normalize_endpoint(self, func):
        return func

    def admin_endpoint_validate(self, func):
        return func

    @abstractmethod
    async def extract_token(self, *args, **kwargs):
        ...

    def admin_endpoint(self, func):
        self.admin_endpoint_validate(func)
        func = self.normalize_endpoint(func)

        @wraps(func)
        async def wrapper(*args, **kwargs):
            token = await self.extract_token(*args, **kwargs)
            self.validate_token(token)
            return await func(*args, **kwargs)
        return wrapper

    # Handlers
    @classmethod
    @abstractmethod
    def unauthorized_handler(cls):
        ...

    @classmethod
    @abstractmethod
    def invalid_admin_credentials_handler(cls):
        ...



from dataclasses import Field, dataclass
from typing import (
    Any,
    NotRequired,
    Tuple,
    Type,
    ClassVar,
    Optional,
    Callable,
    TypedDict,
    Union,
    Literal,
    List
)
import polars as pl
from pydantic import BaseModel, ConfigDict, GetCoreSchemaHandler
from pydantic_core import CoreSchema

from abc import ABC, abstractmethod


@dataclass
class DataConversion:
    name: str
    transform: Callable[[Any], str] = lambda x: f"{x}"


@dataclass
class MethodArgument:
    st: str
    tp: Type


class GlobalTypedMethod_Interface(ABC):
    @abstractmethod
    def __init__(
        self,
        return_type: Type,
        definition: Callable,
        args: dict[str, Union[Optional[Type], list[Optional[Type]]]]
    ) -> None:
        ...

    @abstractmethod
    def check_type(
        self,
        expected_type: Union[Optional[Type], list[Optional[Type]]],
        compared_type: Optional[Type]
    ) -> bool:
        ...

    @abstractmethod
    def __call__(
        self, *args, **kwds
    ) -> Tuple[Any, Type]:
        ...


class TypedMethod_Interface(GlobalTypedMethod_Interface):
    NODE_COMPATIBILITY: dict[Type, Type]

    @abstractmethod
    def __init__(
        self,
        caller_type: Type,
        return_type: Type,
        definition: Callable,
        args: dict[str, Union[Optional[Type], list[Optional[Type]]]]
    ) -> None:
        ...

    @abstractmethod
    def __call__(
        self,
        caller: MethodArgument,
        *args, **kwds
    ) -> Tuple[Any, Type]:
        ...


class DataTransformer_Interface(ABC):
    TYPE_DICT: dict[Type, DataConversion] = {}

    class OperatorsDict(TypedDict):
        numeric: NotRequired[dict[str, str]]
        binary: NotRequired[dict[str, Callable[[Any, Any], str]]]
        boolean: NotRequired[dict[str, str]]
        unary: NotRequired[dict[str, str]]
    OPERATOR_DICT: OperatorsDict

    class MethodsDict(TypedDict):
        globals: dict[str, GlobalTypedMethod_Interface]
        by_type: dict[Type, dict[str, TypedMethod_Interface]]
    METHODS_MAP: MethodsDict

    NODE_COMPATIBILITY: dict[Type, Type]

    @classmethod
    @abstractmethod
    def check_type_compatibility(
        cls,
        tp1: Optional[Type],
        tp2: Optional[Type]
    ) -> bool:
        ...

    @classmethod
    @abstractmethod
    def get_type_from_sql_type(
        cls,
        sql_type: str
    ) -> Type:
        ...

    @classmethod
    @abstractmethod
    def get_method(
        cls,
        caller_type: Type,
        method_name: str
    ) -> Union[GlobalTypedMethod_Interface, TypedMethod_Interface]:
        ...

    @classmethod
    @abstractmethod
    def get_operator(
        cls,
        operator_type: Literal["numeric", "binary", "boolean", "unary"],
        op: Any
    ) -> Union[Callable, str]:
        ...

    @classmethod
    @abstractmethod
    def get_data_field(
        cls,
        v: Any
    ) -> Optional[DataConversion]:
        ...

    @classmethod
    @abstractmethod
    def convert_data_schema(
        cls,
        schema: dict[str, Type]
    ) -> dict[str, str]:
        ...

    @staticmethod
    @abstractmethod
    def validate_data_from_schema(
        schema: dict[str, Type],
        data: dict[str, Any]
    ) -> None:
        ...

    @classmethod
    @abstractmethod
    def convert_data(
        cls,
        data: Any
    ) -> str:
        ...


class QueryBuilder_Interface(ABC):
    @abstractmethod
    def __init__(
        self,
        data_transformer: DataTransformer_Interface
    ) -> None:
        self._data_transformer = data_transformer

    @abstractmethod
    def select(
        self,
        from_table: str,
        columns: List[str],
        where: Optional[str] = None,
        limit: Optional[int] = 10,
        order_by: Optional[str] = None,
        group_by: Optional[list[str]] = None,
        offset: Optional[int] = None
    ) -> str:
        ...

    @abstractmethod
    def insert(
        self,
        into_table: str,
        values: List[dict]
    ) -> str:
        ...

    @abstractmethod
    def update(
        self,
        table: str,
        values: dict,
        where: str
    ) -> str:
        ...

    @abstractmethod
    def delete(
        self,
        table: str,
        where: str
    ) -> str:
        ...

    @abstractmethod
    def create_table(
        self,
        table_name: str,
        schema: dict,
        if_exists: bool = False
    ) -> str:
        ...

    @abstractmethod
    def drop_table(
        self,
        table_name: str,
        if_exists: bool = False
    ) -> str:
        ...


class primary_key_Interface(str, ABC):
    @classmethod
    @abstractmethod
    def generate(cls) -> "primary_key_Interface":
        ...

    @abstractmethod
    def is_valid(self) -> bool:
        ...

    @staticmethod
    @abstractmethod
    def auto_increment() -> Field:
        ...

    @abstractmethod
    def __eq__(
        self,
        value: Any
    ) -> bool:
        ...

    @classmethod
    @abstractmethod
    def __get_pydantic_core_schema__(
        cls,
        source_type: Any,
        handler: GetCoreSchemaHandler
    ) -> CoreSchema:
        ...

    @abstractmethod
    def __hash__(self) -> int:
        ...


class Table_Interface(ABC):
    @abstractmethod
    def __init__(
        self,
        table_name: str,
        schema: dict[str, Type],
        db: Optional["DB_Connection_Interface"] = None
    ) -> None:
        ...

    @property
    @abstractmethod
    def db(self) -> "DB_Connection_Interface":
        ...

    @abstractmethod
    def set_db(
        self,
        db: "DB_Connection_Interface"
    ) -> None:
        ...

    @abstractmethod
    def get_all(
        self,
        limit: Optional[int] = 10,
        only_current: bool = True,
        only_active: bool = True,
        skip: int = 0
    ) -> Optional[pl.DataFrame]:
        ...

    @abstractmethod
    def filter(
        self,
        l_func: Callable[[Any], bool],
        limit: Optional[int] = 10,
        only_current: bool = True,
        only_active: bool = True,
        skip: int = 0
    ) -> Optional[pl.DataFrame]:
        ...

    @abstractmethod
    def insert(
        self,
        data: dict[str, Any]
    ) -> None:
        ...

    @abstractmethod
    def update(
        self,
        l_func: Callable[[Any], bool],
        data: dict[str, Any]
    ) -> None:
        ...

    @abstractmethod
    def delete_all(self) -> None:
        ...


class Object_Interface(BaseModel, ABC):
    class DependencyRequiredIds(BaseModel):
        type: Type['Object_Interface']
        ids: set[str]

    class ObjectDependency(BaseModel):
        model_config = ConfigDict(arbitrary_types_allowed=True)
        is_list: bool
        is_value: bool
        type: Type['Object_Interface']
        table: 'Table_Interface'
    __table: ClassVar[Optional['Table_Interface']]
    __dependencies: ClassVar[dict[str, ObjectDependency]]
    identifier_field: ClassVar[str]
    __objects_list: ClassVar[List['Object_Interface']]
    __objects_map: ClassVar[dict[str, 'Object_Interface']]

    @classmethod
    @abstractmethod
    def delete_all(cls) -> None:
        ...

    @classmethod
    @abstractmethod
    def set_db(
        cls,
        db: "DB_Connection_Interface"
    ) -> None:
        ...

    class TypeDictionary(TypedDict):
        default: NotRequired[Any]
        default_factory: NotRequired[Optional[Callable]]
        is_list: NotRequired[bool]
        nullable: NotRequired[bool]
        type: NotRequired[Optional[Type]]
        depends: NotRequired[Type]

    @classmethod
    @abstractmethod
    def get_types(cls) -> dict[str, TypeDictionary]:
        ...

    @classmethod
    @abstractmethod
    def _update_individual(
        cls,
        data_inp: dict[str, Any]
    ) -> Optional['Object_Interface']:
        ...

    @classmethod
    @abstractmethod
    def _update_info(
        cls,
        filter: Optional[Callable[[Any], bool]] = None,
        limit: Optional[int] = 10,
        skip: int = 0,
        only_current: bool = True,
        only_active: bool = True
    ) -> list['Object_Interface']:
        ...

    @classmethod
    @abstractmethod
    def new(cls, **kwargs) -> 'Object_Interface':
        ...

    @abstractmethod
    def history(self) -> dict[str, list[dict[str, Any]]]:
        ...

    @abstractmethod
    def update(self, **kwargs) -> None:
        ...

    @abstractmethod
    def delete(self) -> None:
        ...

    @classmethod
    @abstractmethod
    def all(
        cls,
        limit: Optional[int] = 10,
        skip: int = 0
    ) -> list["Object_Interface"]:
        ...

    @classmethod
    @abstractmethod
    def filter(
        cls,
        lambda_f: Optional[Callable[[Any], bool]],
        limit: Optional[int] = 10,
        skip: int = 0
    ) -> list["Object_Interface"]:
        ...

    @classmethod
    @abstractmethod
    def get_by_id(
        cls,
        id_param: str
    ) -> Optional["Object_Interface"]:
        ...

    @classmethod
    @abstractmethod
    def get_table_res(
        cls,
        limit: int = 10,
        skip: int = 0,
        only_current: bool = True,
        only_active: bool = True
    ) -> Optional[pl.DataFrame]:
        ...

    @abstractmethod
    def to_dict(self) -> dict[str, Any]:
        ...

    @abstractmethod
    def to_json(self) -> str:
        ...

    @abstractmethod
    def __getitem__(self, item: str) -> Any:
        ...


class TableName(BaseModel):
    name: str
    alias: str


class TableField(TypedDict):
    primary_key: NotRequired[bool]
    type: str
    is_list: NotRequired[bool]
    nullable: NotRequired[bool]
    depends: NotRequired[str]


class DB_Connection_Interface(ABC):
    __data_transformer: DataTransformer_Interface
    __query: QueryBuilder_Interface
    __classes: dict[str, Type[Object_Interface]] = {}
    _table_schemas: dict[str, dict[str, Type]] = {}
    _tables: list[str] = []

    @property
    @abstractmethod
    def usage_metrics(self) -> dict[str, int]:
        ...

    @abstractmethod
    def exit(self) -> None:
        ...

    @abstractmethod
    def __init__(
        self,
        data_transformer: DataTransformer_Interface,
        query: QueryBuilder_Interface
    ) -> None:
        ...

    @property
    @abstractmethod
    def query(self) -> QueryBuilder_Interface:
        ...

    @property
    @abstractmethod
    def data_transformer(self) -> DataTransformer_Interface:
        ...

    @abstractmethod
    def get_table_name(
        self,
        table_name: str
    ) -> TableName:
        ...

    @abstractmethod
    def clean_db(
        self,
        DO_NOT_ASK: bool = False
    ) -> None:
        ...

    @abstractmethod
    def get_table_definition(
        self,
        table_name: str
    ) -> dict[str, Type]:
        ...

    @abstractmethod
    def attach(
        self,
        objects: list[Type[Object_Interface]]
    ) -> None:
        ...

    @abstractmethod
    def get_json_schema(self) -> dict[str, dict[str, TableField]]:
        ...

    @property
    @abstractmethod
    def classes(self) -> list[Type["Object_Interface"]]:
        ...

    @abstractmethod
    def execute(
        self,
        query: str
    ) -> Optional[pl.DataFrame]:
        ...

    @abstractmethod
    def normalize_statement(
        self,
        statement: str
    ) -> str:
        ...

    @abstractmethod
    def ensure_table(
        self,
        table_name: str,
        schema: dict[str, Type],
        save: bool = False,
        current_data_schema: Optional[dict[str, Type]] = None
    ) -> None:
        ...



import polars as pl
from typing import Optional, Type, get_origin, get_args
from autodla.utils.logger import logger

from autodla.engine.interfaces import (
    DB_Connection_Interface,
    TableName,
    Object_Interface,
    QueryBuilder_Interface,
    DataTransformer_Interface,
    TableField
)


class DB_Connection(DB_Connection_Interface):
    __data_transformer: DataTransformer_Interface
    __query: QueryBuilder_Interface
    __classes: dict[str, Type[Object_Interface]] = {}
    _table_schemas: dict[str, dict[str, Type]] = {}
    _tables: list[str] = []

    @property
    def usage_metrics(self) -> dict[str, int]:
        """
        Returns the usage metrics of the database connection.
        This is a placeholder method and should be implemented in subclasses.
        """
        raise NotImplementedError(
            "This method should be implemented in subclasses."
        )

    def exit(self) -> None:
        pass

    def __init__(
        self,
        data_transformer: DataTransformer_Interface,
        query: QueryBuilder_Interface
    ) -> None:
        self.__data_transformer = data_transformer
        self.__query = query
        self.__classes: dict[str, Type[Object_Interface]] = {}
        self._table_schemas: dict[str, dict[str, Type]] = {}
        self._tables: list[str] = []

    @property
    def query(self) -> QueryBuilder_Interface:
        return self.__query

    @property
    def data_transformer(self) -> DataTransformer_Interface:
        return self.__data_transformer

    def get_table_name(
        self,
        table_name: str
    ) -> TableName:
        """
        Returns the table name in uppercase.
        This is used to ensure that the table names
        are consistent with the SQL standard.
        """
        return TableName(name=table_name.upper(), alias=table_name.lower())

    def clean_db(
        self,
        DO_NOT_ASK: bool = False
    ) -> None:
        if not DO_NOT_ASK:
            logger.debug("Are you sure you want to clean the database? (y/n)")
            answer = input()
            if answer != "y":
                raise Exception("User did not confirm the action")
        logger.debug("Cleaning database...")
        for class_i in self.__classes.values():
            class_i.delete_all()
        logger.debug("Database cleaned")

    def get_table_definition(
        self,
        table_name: str
    ) -> dict[str, type]:
        return {}

    def attach(
        self,
        objects: list[Type[Object_Interface]]
    ) -> None:
        ordered_objects = []
        pending = objects[:]
        while True:
            if pending == []:
                break
            tmp = pending[:]
            for obj in tmp:
                schema = obj.get_types()
                class_dependencies = []
                for i in schema.values():
                    if 'depends' in i:
                        class_dependencies.append(i.get('depends'))
                if all([dep in ordered_objects for dep in class_dependencies]):
                    ordered_objects.append(obj)
                    pending.remove(obj)
        for obj in ordered_objects:
            self.__classes[obj.__name__] = obj
            obj.set_db(self)

    def get_json_schema(self) -> dict[str, dict[str, TableField]]:
        out: dict[str, dict[str, TableField]] = {}
        for class_key, class_i in self.__classes.items():
            class_def = class_i.get_types()
            class_out: dict[str, TableField] = {}
            for k, f in class_def.items():
                if f["type"] is None:
                    continue
                type_st = f["type"].__name__
                if get_origin(f['type']) == list:
                    arg = get_args(f["type"])
                    if len(arg) == 1:
                        type_st += f'[{arg[0].__name__}]'
                class_out[k] = {
                    "type": type_st
                }
                if class_i.identifier_field == k:
                    class_out[k]["primary_key"] = True
                if "depends" in f:
                    class_out[k]["depends"] = f'$ref:{f["depends"].__name__}'
                if "is_list" in f:
                    class_out[k]["is_list"] = f["is_list"]
                if "nullable" in f:
                    class_out[k]["nullable"] = f["nullable"]
            out[class_key] = class_out
        return out

    @property
    def classes(self) -> list[Type["Object_Interface"]]:
        return list(self.__classes.values())

    def execute(
        self,
        query: str
    ) -> Optional[pl.DataFrame]:
        return pl.DataFrame()

    def normalize_statement(
        self,
        statement: str
    ) -> str:
        if not isinstance(statement, str):
            statement = str(statement)
        statement = statement.lstrip().rstrip()
        if statement[-1] != ";":
            statement += ";"
        return statement

    def ensure_table(
        self,
        table_name: str,
        schema: dict[str, Type],
        save: bool = False,
        current_data_schema: Optional[dict[str, Type]] = None
    ) -> None:
        logger.debug("ENSURE TABLE DBCONNECTION")
        if save:
            self._table_schemas[table_name] = schema
            self._tables.append(table_name)
        data_schema = {k.upper(): v["type"] for k, v in schema.items()}
        if current_data_schema is None:
            current_data_schema = self.get_table_definition(table_name)
        if all([
            self.data_transformer.check_type_compatibility(
                data_schema.get(k), current_data_schema.get(k)
            ) for k in list(
                set(data_schema.keys()).union(set(data_schema.keys()))
            )
        ]):
            return
        logger.debug(data_schema)
        logger.debug(current_data_schema)
        if data_schema == current_data_schema:
            return
        converted_schema = self.data_transformer.convert_data_schema(schema)
        table_name_db = self.get_table_name(table_name)
        self.execute(self.query.drop_table(table_name_db.name, if_exists=True))
        qry = self.query.create_table(table_name_db.name, converted_schema)
        self.execute(qry)



import pytest
from autodla import Object, primary_key
from autodla.dbs import MemoryDB
from autodla.utils import DataGenerator

class User(Object):
    id: primary_key = primary_key.auto_increment()
    name: str
    age: int

class Team(Object):
    id: primary_key = primary_key.auto_increment()
    participants: list[User]
    group_name: str

class Item(Object):
    id: primary_key = primary_key.auto_increment()
    name: str
    tags: list[str]

@pytest.fixture
def db(monkeypatch):
    from autodla.engine import object as engine_object
    from autodla.engine.lambda_conversion import lambda_to_sql
    original_init = engine_object.Table.__init__
    original_update = engine_object.Table.update

    def init(self, table_name: str, schema: dict, db: MemoryDB | None = None):
        self.table_name = table_name
        self.schema = schema
        if db:
            engine_object.Table.set_db(self, db)

    def tbl_update(self, l_func, data):
        alias = "".join(self.table_name.split("."))
        where_st = lambda_to_sql(self.schema, l_func, self.db.data_transformer, alias=alias)
        update_data = {f'{key}': value for key, value in data.items()}
        qry = self.db.query.update(self.table_name, where=where_st, values=update_data)
        return self.db.execute(qry)

    monkeypatch.setattr(engine_object.Table, "__init__", init)
    monkeypatch.setattr(engine_object.Table, "update", tbl_update)
    db = MemoryDB()
    db.attach([User, Team, Item])
    yield db
    User.delete_all()
    Team.delete_all()
    Item.delete_all()
    monkeypatch.setattr(engine_object.Table, "__init__", original_init)
    monkeypatch.setattr(engine_object.Table, "update", original_update)


def test_create_and_retrieve_user(db):
    user = User.new(name="Alice", age=25)
    assert isinstance(user.id, primary_key)
    users = User.all(limit=None)
    assert len(users) == 1
    assert users[0] is user


def test_filter_users(db):
    u1 = User.new(name="A", age=20)
    u2 = User.new(name="B", age=30)
    res = User.filter(lambda x: x.age >= 25, limit=None)
    assert res == [u2]


def test_update_and_delete_user(db):
    user = User.new(name="John", age=20)
    user.update(age=21)
    assert User.get_by_id(user.id).age == 21
    user.delete()
    assert User.all(limit=None) == []


def test_group_relationship(db):
    u1 = User.new(name=DataGenerator.name(), age=DataGenerator.age())
    grp = Team.new(participants=[u1], group_name="Group1")
    groups = Team.all(limit=None)
    assert groups[0].participants[0] is u1


def test_history_tracks_updates(db):
    user = User.new(name="Hist", age=20)
    user.update(age=21)
    history = user.history()
    assert len(history["self"]) == 2
    operations = [row["DLA_operation"] for row in history["self"]]
    assert operations[0] == "INSERT" and operations[1] == "UPDATE"


def test_get_table_res_only_current(db):
    user = User.new(name="Table", age=22)
    user.update(age=23)
    current = User.get_table_res(limit=None)
    full = User.get_table_res(limit=None, only_current=False, only_active=False)
    assert len(current.to_dicts()) == 1
    assert len(full.to_dicts()) == 2


def test_list_value_dependency(db):
    item = Item.new(name="Item1", tags=["a", "b"])
    item.update(tags=["a", "c"])
    assert item.tags == ["a", "c"]
    hist = item.history()
    assert len(hist["self"]) == 2
    assert len(hist["dependencies"]["tags"]) == 4


def test_delete_preserves_history(db):
    user = User.new(name="Del", age=40)
    user.delete()
    assert User.all(limit=None) == []
    hist = user.history()
    assert len(hist["self"]) == 2
    assert hist["self"][1]["DLA_operation"] == "DELETE"
    assert hist["self"][1]["DLA_is_active"] == 0


def test_complex_filtering(db):
    User.new(name="Alice", age=32)
    User.new(name="Bob", age=26)
    User.new(name="Charlie", age=40)
    res = User.filter(lambda x: (x.age >= 30 or x.age <= 20) and (x.age != 26), limit=None)
    assert len(res) == 2
    names = sorted([u.name for u in res])
    assert names == ["Alice", "Charlie"]


def test_object_identity_consistency(db):
    user = User.new(name="Ident", age=50)
    from_all = User.all(limit=None)[0]
    by_id = User.get_by_id(user.id)
    assert from_all is user
    assert by_id is user


def test_team_update_participants_history(db):
    u1 = User.new(name="U1", age=20)
    u2 = User.new(name="U2", age=22)
    team = Team.new(participants=[u1], group_name="G1")
    team.update(participants=[u1, u2])
    assert team.participants == [u1, u2]
    hist = team.history()
    assert len(hist["self"]) == 2
    assert len(hist["dependencies"]["participants"]) == 3



import os
import uuid
from datetime import datetime

import psycopg2

from autodla import Object, primary_key
from autodla.dbs import PostgresDB


def test_postgres_connection(monkeypatch):
    os.environ.setdefault("AUTODLA_POSTGRES_USER", "postgres")
    os.environ.setdefault("AUTODLA_POSTGRES_PASSWORD", "password")
    os.environ.setdefault("AUTODLA_POSTGRES_HOST", "localhost")
    os.environ.setdefault("AUTODLA_POSTGRES_DB", "my_db")

    # verify raw connection works
    conn = psycopg2.connect(
        user=os.environ["AUTODLA_POSTGRES_USER"],
        password=os.environ["AUTODLA_POSTGRES_PASSWORD"],
        host=os.environ["AUTODLA_POSTGRES_HOST"],
        dbname=os.environ["AUTODLA_POSTGRES_DB"],
    )
    with conn.cursor() as cur:
        cur.execute("SELECT 1")
        assert cur.fetchone()[0] == 1
    conn.close()

    from autodla.engine import object as engine_object
    from autodla.engine.lambda_conversion import lambda_to_sql

    original_init = engine_object.Table.__init__
    original_update = engine_object.Table.update

    def init(self, table_name: str, schema: dict, db: PostgresDB | None = None):
        self.table_name = table_name
        self.schema = schema
        if db:
            engine_object.Table.set_db(self, db)

    def tbl_update(self, l_func, data):
        alias = "".join(self.table_name.split("."))
        where_st = lambda_to_sql(self.schema, l_func, self.db.data_transformer, alias=alias)
        update_data = {f"{key}": value for key, value in data.items()}
        qry = self.db.query.update(self.table_name, where=where_st, values=update_data)
        return self.db.execute(qry)

    monkeypatch.setattr(engine_object.Table, "__init__", init)
    monkeypatch.setattr(engine_object.Table, "update", tbl_update)

    class Person(Object):
        id: primary_key = primary_key.auto_increment()
        name: str

    db = PostgresDB(sync_interval=3600)
    db.attach([Person])

    pg_conn = db._PostgresDB__pg_connection
    pg_query = db._PostgresDB__pg_query
    pg_dt = db._PostgresDB__pg_dt

    schema = Person._Object__table.schema.copy()
    schema["DLA_is_current"]["type"] = int
    schema["DLA_is_active"]["type"] = int
    with pg_conn.cursor() as cur:
        cur.execute(pg_query.drop_table("person", if_exists=True))
        cur.execute(pg_query.create_table("person", pg_dt.convert_data_schema(schema)))
    pg_conn.commit()

    try:
        p1 = Person.new(name="Alice")
        db.synchronize()
        with pg_conn.cursor() as cur:
            cur.execute("SELECT name FROM person")
            rows = cur.fetchall()
        assert rows == [("Alice",)]

        with pg_conn.cursor() as cur:
            cur.execute(
                "INSERT INTO person (id, name, DLA_object_id, DLA_modified_at, DLA_operation, DLA_modified_by, DLA_is_current, DLA_is_active)"
                " VALUES (%s, %s, %s, %s, %s, %s, %s, %s)",
                (
                    str(uuid.uuid4()),
                    "Bob",
                    str(uuid.uuid4()),
                    datetime.now(),
                    "INSERT",
                    "SYSTEM",
                    1,
                    1,
                ),
            )
        pg_conn.commit()

        db.synchronize()
        names = sorted([p.name for p in Person.all(limit=None)])
        assert names == ["Alice", "Bob"]

        p1.update(name="Alice2")
        db.synchronize()

        with pg_conn.cursor() as cur:
            cur.execute(
                "SELECT name, DLA_operation, DLA_is_current FROM person ORDER BY DLA_modified_at"
            )
            rows = cur.fetchall()

        assert ("Alice", "INSERT", 0) in rows
        assert ("Bob", "INSERT", 1) in rows
        assert ("Alice2", "UPDATE", 1) in rows
    finally:
        with pg_conn.cursor() as cur:
            cur.execute(pg_query.drop_table("person", if_exists=True))
        pg_conn.commit()

    db._PostgresDB__pg_connection.close()



# The `DataTransformer` class in the provided Python code contains
# methods for data conversion, type checking, method retrieval,
# and SQL operator mapping.
from typing import (
    Callable,
    Any,
    Literal,
    Optional,
    Tuple,
    Type,
    Union
)
from autodla.engine.interfaces import (
    DataConversion,
    MethodArgument,
    GlobalTypedMethod_Interface,
    TypedMethod_Interface,
    DataTransformer_Interface
)
import os

DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'
if "DATETIME_FORMAT" in os.environ:
    DATETIME_FORMAT = os.environ.get("DATETIME_FORMAT", "")


class GlobalTypedMethod(GlobalTypedMethod_Interface):
    def __init__(
        self,
        return_type: Type,
        definition: Callable,
        args: dict[str, Union[Optional[Type], list[Optional[Type]]]]
    ) -> None:
        self.return_type = return_type
        self.definition = definition
        self.args = args

    def check_type(
        self,
        expected_type: Union[Optional[Type], list[Optional[Type]]],
        compared_type: Optional[Type]
    ) -> bool:
        if type(expected_type) is list:
            for type_i in expected_type:
                if type_i is None and compared_type is None:
                    return True
                if compared_type == type_i:
                    return True
        elif compared_type == expected_type:
            return True
        return False

    def __call__(
        self, *args, **kwds
    ) -> Tuple[Any, Type]:
        keys = list(self.args.keys())
        arguments = {}
        for i, k in enumerate(keys):
            v = None
            if i < len(args):
                v = args[i]
            if k in kwds:
                v = kwds[k]
            if not self.check_type(self.args[k], getattr(v, 'tp', None)):
                raise TypeError(
                    "expected type {} for argument {}, got {}".format(
                        self.args[k],
                        k,
                        type(v)
                    )
                )
            arguments[k] = getattr(v, 'st', None)
        return self.definition(arguments), self.return_type


class TypedMethod(GlobalTypedMethod, TypedMethod_Interface):
    def __init__(
        self,
        caller_type: Type,
        return_type: Type,
        definition: Callable,
        args: dict[str, Union[Optional[Type], list[Optional[Type]]]]
    ) -> None:
        self.caller_type = caller_type
        super().__init__(return_type, definition, args)

    def __call__(
        self,
        caller: MethodArgument,
        *args, **kwds
    ) -> Tuple[Any, Type]:
        if (
            any([(
                type(arg) is not MethodArgument
            ) for arg in args])
            or any([(
                type(arg) is not MethodArgument
            ) for arg in list(kwds.values())])
        ):
            raise ValueError(
                "All arguments for data conversion should be MethodArgument"
            )
        keys = list(self.args.keys())
        arguments = {}
        for i, k in enumerate(keys):
            v: Optional[MethodArgument] = None
            if i < len(args):
                v = args[i]
            if k in kwds:
                v = kwds[k]
            if v is None:
                raise ValueError(f"Argument {k} is required")
            if not self.check_type(
                self.args[k], v.tp
            ):
                raise TypeError(
                    "expected type {} for argument {}, got {}".format(
                        self.args[k],
                        k,
                        v.tp
                    )
                )
            arguments[k] = v.st
        if not self.check_type(self.caller_type, caller.tp):
            raise TypeError(
                "expected caller type {}, got {}".format(
                    self.caller_type,
                    caller.tp
                )
            )
        arguments["caller"] = caller.st
        return self.definition(arguments), self.return_type


class DataTransformer(DataTransformer_Interface):
    TYPE_DICT: dict[Type, DataConversion]
    OPERATOR_DICT: DataTransformer_Interface.OperatorsDict
    METHODS_MAP: DataTransformer_Interface.MethodsDict
    NODE_COMPATIBILITY: dict[Type, Type] = {}

    @classmethod
    def check_type_compatibility(cls, tp1, tp2) -> bool:
        out = tp1 == tp2
        if not out:
            if tp1 in cls.NODE_COMPATIBILITY:
                out = cls.NODE_COMPATIBILITY[tp1] == tp2
        return out

    @classmethod
    def get_type_from_sql_type(cls, sql_type) -> type:
        found = None
        for k, v in cls.TYPE_DICT.items():
            if v.name.upper() == sql_type.upper():
                found = k
        if found is None:
            raise TypeError(f"invalid conversion for sql_type '{sql_type}'")
        return found

    @classmethod
    def get_method(
        cls,
        caller_type: Type,
        method_name: str
    ) -> Union[GlobalTypedMethod_Interface, TypedMethod_Interface]:
        d: Union[
            dict[str, GlobalTypedMethod_Interface],
            dict[str, TypedMethod_Interface]
        ] = cls.METHODS_MAP["globals"]
        if caller_type is not None:
            d = cls.METHODS_MAP["by_type"][caller_type]
        out = d.get(method_name)
        if out is None:
            raise ValueError(
                "Method '{}' not found for type {}".format(
                    method_name,
                    caller_type.__name__ if caller_type else 'None'
                )
            )
        return out

    @classmethod
    def get_operator(
        cls,
        operator_type: Literal["numeric", "binary", "boolean", "unary"],
        op: Any
    ) -> Union[Callable, str]:
        out = cls.OPERATOR_DICT[operator_type].get(type(op).__name__)
        if out is None:
            raise ValueError(
                "Unsupported {} operator: {}".format(
                    operator_type,
                    op.__class__.__name__
                )
            )
        return out

    @classmethod
    def get_data_field(
        cls,
        v: Any
    ) -> Optional[DataConversion]:
        return cls.TYPE_DICT.get(v)

    @classmethod
    def convert_data_schema(
        cls,
        schema: dict[str, Type]
    ) -> dict[str, str]:
        out = {}
        for k, v in schema.items():
            f = cls.get_data_field(v["type"])
            if f is not None:
                out[k] = f.name
                if v.get("nullable") is not True:
                    out[k] += " NOT NULL"
        return out

    @staticmethod
    def validate_data_from_schema(
        schema: dict[str, Type],
        data: dict[str, Any]
    ) -> None:
        extra_keys = []
        for i in data:
            if i not in schema:
                extra_keys.append(i)
        if extra_keys:
            raise ValueError(f"Extra values found: {extra_keys}")
        missing_values = []
        for i in data:
            if i not in schema:
                missing_values.append(i)
        if missing_values:
            raise ValueError(f"Missing values: {missing_values}")
        invalid_types: list[str] = []
        for i in data:
            if not isinstance(data[i], schema[i]):
                invalid_types.append(i)
        if invalid_types:
            msg = 'Invalid types:\n'
            for invalid_type in invalid_types:
                msg += '{}: expected {} got {}\n'.format(
                    invalid_type,
                    schema[invalid_type],
                    type(data[invalid_type])
                )
            raise ValueError(msg)

    @classmethod
    def convert_data(
        cls,
        data: Any
    ) -> str:
        v = cls.get_data_field(type(data))
        if v is not None:
            return v.transform(data)
        if type(data) is list:
            return f"({', '.join([cls.convert_data(i) for i in data])})"
        raise TypeError(
            f"Missing transformer for class {type(data).__name__}"
        )
